def to_readable_name(key: str) -> str:
    """
    Converts prompt_toolkit key names to human-readable names (for more info see https://python-prompt-toolkit.readthedocs.io/en/master/pages/advanced_topics/key_bindings.html#list-of-special-keys)
    ie:
        - "c-p" -> "CTRL+P"
        - "c-x" -> "CTRL+X"
        - "enter" -> "ENTER"
        - etc

    :param key: the key to convert
    :return: the human-readable key
    """
    if key.startswith("c-"):
        return "CTRL+" + key[2:].upper()
    elif key.startswith("s-"):
        return "SHIFT+" + key[2:].upper()
    elif key.startswith("a-"):
        return "ALT+" + key[2:].upper()
    elif key in ["escape", "enter", "space", "tab", "backspace"]:
        return key.upper()
    elif key.startswith("up"):
        return "↑"
    elif key.startswith("down"):
        return "↓"
    elif key.startswith("left"):
        return "←"
    elif key.startswith("right"):
        return "→"
    else:
        raise ValueError(f"Unknown key {key}")



from typing import List, Callable


class HexagonError(Exception):
    """Base class for all Hexagon errors."""

    def __init__(self, error_printer: Callable, exit_status: int = 1) -> None:
        super().__init__("hexagon error")
        self.exit_status = exit_status
        self.error_printer = error_printer

    def print_error(self, logger):
        self.error_printer(logger)


class ListHexagonError(HexagonError):
    """A list of errors."""

    def __init__(self, errors: List[str]) -> None:
        self.errors = errors
        super().__init__(self._error_printer)

    def _error_printer(self, logger) -> None:
        for error in self.errors:
            logger.error(error)


class SilentHexagonError(HexagonError):
    """A silent error."""

    def __init__(self) -> None:
        super().__init__(lambda x: None)



from enum import Enum, auto
from pathlib import Path


class InquiryType(Enum):
    STRING = auto()
    STRING_SEARCHABLE = auto()
    STRING_LIST = auto()
    STRING_LIST_SEARCHABLE = auto()
    ENUM = auto()
    ENUM_SEARCHABLE = auto()
    ENUM_LIST = auto()
    ENUM_LIST_SEARCHABLE = auto()
    PATH = auto()
    PATH_SEARCHABLE = auto()
    INT = auto()
    FLOAT = auto()
    SECRET = auto()
    BOOLEAN = auto()


def determine_expected_inquiry(iterable, of_enum, field_type, extras) -> InquiryType:
    searchable = extras.get("searchable", False)

    query = InquiryType.STRING if not searchable else InquiryType.STRING_SEARCHABLE
    if iterable and of_enum:
        query = (
            InquiryType.ENUM_LIST
            if not searchable
            else InquiryType.ENUM_LIST_SEARCHABLE
        )
    elif iterable:
        query = (
            InquiryType.STRING_LIST
            if not searchable
            else InquiryType.STRING_LIST_SEARCHABLE
        )
    elif of_enum:
        query = InquiryType.ENUM if not searchable else InquiryType.ENUM_SEARCHABLE
    elif issubclass(field_type.base_type, Path):
        query = InquiryType.PATH if not searchable else InquiryType.PATH_SEARCHABLE
    elif issubclass(field_type.base_type, bool):
        query = InquiryType.BOOLEAN
    elif issubclass(field_type.base_type, int):
        query = InquiryType.INT
    elif issubclass(field_type.base_type, float):
        query = InquiryType.FLOAT
    elif extras.get("secret", False):
        query = InquiryType.SECRET

    return query



import threading
from enum import Enum
from typing import Callable, Generic, List, TypeVar


class HookSubscrptionType(Enum):
    blocking = "blocking"
    background = "background"


T = TypeVar("T")


class HookSubscription(Generic[T]):
    def __init__(
        self,
        name: str,
        callback: Callable[[T], None],
        type: HookSubscrptionType = HookSubscrptionType.blocking,
    ):
        self.name = name
        self.type = type
        self.callback = callback

    callback: Callable[[T], None]


class Hook(Generic[T]):
    def __init__(self, name: str):
        self.name = name
        self.subscriptions = []

    subscriptions: List[HookSubscription[T]]

    def subscribe(self, subscription: HookSubscription[T]):
        self.subscriptions.append(subscription)

    def run(self, data: T = None):
        for subscription in self.subscriptions:
            if subscription.type == HookSubscrptionType.blocking:
                subscription.callback(data)
            elif subscription.type == HookSubscrptionType.background:
                thread = threading.Thread(
                    target=subscription.callback,
                    name=f"{self.name}:{subscription.name}",
                    kwargs={"data": data},
                )
                thread.daemon = True
                thread.start()
            else:
                # Unknown HookSubscriptionType {subscription_type}
                raise Exception(
                    _("error.support.hooks.hook.unknown_type").format(
                        subscription_type=subscription.type
                    )
                )



from typing import Any, Optional

from pydantic import validator

from hexagon.domain.env import Env
from hexagon.domain.tool import ActionTool
from hexagon.support.input.args import ToolArgs, Arg, OptionalArg
from hexagon.support.output.printer import log


class Args(ToolArgs):
    last_name: OptionalArg[str] = Arg(
        None,
        prompt_message="What's your last name?",
    )

    @validator("last_name")
    def not_empty(cls, arg):
        if not arg.value:
            raise ValueError("Please enter your last name.")
        return arg


def main(
    tool: ActionTool,
    env: Optional[Env] = None,
    env_args: Any = None,
    cli_args: Args = None,
):
    """
    All hexagon tools must define a main function

    :param tool: the ActionTool that is being executed, ie: all values present in the yaml
    :param env: the Env that is being executed, ie: all values present in the yaml
    :param env_args: the values expected to receive from tool.envs[env.name]
    :param cli_args: any extra arguments passed when executing hexagon, ie: --foo "bar"
    :return:
    """

    # the ideal way of asking the user for input is to use cli_args.prompt
    # this will make sure to trace the input accordingly
    if not cli_args.last_name:
        cli_args.last_name.prompt()

    log.info(f"selected tool: {tool.action}")
    log.info(f"selected env: {env}")
    log.info(f"values in tool.envs\\[env.name]: {env_args}")
    log.info(f"extra cli arguments: {cli_args}")
    log.result(f"your last name is: {cli_args.last_name}")



import os
from pathlib import Path

from pydantic import FilePath, DirectoryPath, field_validator

from hexagon.runtime.dependencies import scan_and_install_dependencies
from hexagon.runtime.singletons import configuration
from hexagon.support.input.args import ToolArgs, Arg, PositionalArg, OptionalArg
from hexagon.support.output.printer import log
from hexagon.support.storage import (
    load_user_data,
    HexagonStorageKeys,
    store_user_data,
)


class Args(ToolArgs):
    src_path: PositionalArg[FilePath] = Arg(
        None,
        prompt_message=_("action.actions.internal.install_cli.config_file_location"),
    )
    bin_path: OptionalArg[DirectoryPath] = Arg(
        None,
        prompt_message=_("action.actions.internal.install_cli.commands_path"),
        prompt_default=str(os.path.expanduser(os.path.join("~", ".local", "bin"))),
    )

    @field_validator("src_path")
    @classmethod
    def is_yaml(cls, arg: PositionalArg[FilePath]):
        if arg.value is None:
            return arg
        if arg.value.suffix == ".yaml" or arg.value.suffix == ".yml":
            return arg
        raise ValueError(_("error.actions.internal.install_cli.select_valid_file"))


def main(_tool, _env, _env_args, cli_args: Args):
    if not cli_args.src_path.value:
        cli_args.src_path.prompt(default=str(Path.cwd()))

    cli, tools, envs = configuration.init_config(cli_args.src_path.value.resolve())

    bin_path = (
        load_user_data(HexagonStorageKeys.cli_install_path.value)
        or cli_args.bin_path.value
    )

    if not bin_path:
        if not cli_args.bin_path.value:
            bin_path = cli_args.bin_path.prompt().resolve()
        store_user_data(HexagonStorageKeys.cli_install_path.value, str(bin_path))

    command_path = os.path.join(bin_path, cli.command)
    with open(command_path, "w") as command:
        pre_command = (
            cli.entrypoint.pre_command + " " if cli.entrypoint.pre_command else ""
        )
        command.write(
            "\n".join(
                [
                    f"#!{cli.entrypoint.shell or '/bin/bash'}",
                    "# file created by hexagon",
                    f"HEXAGON_CONFIG_FILE={cli_args.src_path.value.resolve()} \\",
                    *[
                        f"{key}={value} \\"
                        for key, value in cli.entrypoint.environ.items()
                    ],
                    f"{pre_command}hexagon $@",
                ]
            )
        )

    _make_executable(command_path)
    scan_and_install_dependencies(configuration.custom_tools_path)

    log.info(
        _("msg.actions.internal.install_cli.success"),
        gap_end=1,
        gap_start=1,
    )
    log.result(f"[b]$ {cli.command}")

    path = os.getenv("PATH").split(":")

    if str(bin_path) not in path:
        log.info(
            _("msg.actions.internal.install_cli.not_in_path").format(dir=bin_path),
            gap_start=1,
        )


def _make_executable(path):
    mode = os.stat(path).st_mode
    mode |= (mode & 0o444) >> 2  # copy R bits to X
    os.chmod(path, mode)



import os
import sys
import traceback
from pathlib import Path
from typing import Optional

from pydantic import ValidationError
from rich import traceback as rich_traceback

from hexagon.domain.hexagon_error import ListHexagonError, HexagonError
from hexagon.support.output.printer import Logger


class ToolExecutionError(ListHexagonError):
    def __init__(
        self, return_code, executed_command, executable_str, custom_tools_path
    ) -> None:
        errors = [
            _("error.support.execute.errors.command_result_code").format(
                executed_command=executed_command, return_code=return_code
            )
        ]

        if return_code == 127:
            errors.append(
                _("error.support.execute.errors.could_not_execute").format(
                    action=executable_str
                )
            )
            errors.append(_("error.support.execute.errors.we_tried"))
            errors.append(
                _("error.support.execute.errors.attempt_cli_custom_dir").format(
                    path=custom_tools_path
                )
            )
            errors.append(
                _("error.support.execute.errors.attempt_internal_tools").format(
                    package="hexagon.actions.external"
                )
            )
            errors.append(
                _("error.support.execute.errors.attempt_known_script").format(
                    extensions=".js, .sh"
                )
            )
            errors.append(_("error.support.execute.errors.attempt_inline_command"))
        super().__init__(errors)


class ActionInputError(ListHexagonError):
    def __init__(self, e: ValidationError, tool_name: str) -> None:
        errors = [
            _("error.support.execute.errors.invalid_input").format(
                count=len(e.errors()), tool=tool_name
            )
        ]
        for error in e.errors():
            errors.append(
                os.linesep
                + _("error.support.execute.errors.invalid_argument").format(
                    loc=error["loc"][0], message=error["msg"]
                )
            )
        super().__init__(errors)


class WithTracebackError(HexagonError):
    def __init__(self, action_id: str, custom_tools_path: Optional[str] = None) -> None:
        tb, execution = self.__pretty_print_external_error(action_id, custom_tools_path)
        self.traceback = tb
        self.execution = execution
        self.error = None
        super().__init__(self._error_printer)

    def _error_printer(self, logger: Logger):
        if self.traceback:
            logger.example(self.traceback, decorator_start=False, decorator_end=False)
        else:
            logger.error(self.execution)

        logger.error(self.error)

    @classmethod
    def __pretty_print_external_error(
        cls, action_id: str, custom_tools_path: Optional[str]
    ) -> (str, str):
        exc_type, exc_value, tb = sys.exc_info()

        trace = (
            cls.__find_python_module_in_traceback(action_id, tb, custom_tools_path)
            if custom_tools_path
            else tb
        )

        return (
            rich_traceback.Traceback.from_exception(exc_type, exc_value, trace),
            exc_value,
        )

    @classmethod
    def __find_python_module_in_traceback(
        cls, action_id: str, tb, custom_tools_path: str
    ):
        return next(
            (
                t
                for t, path, file_name in cls.__walk_tb(tb)
                if cls.__executed_action_in_traceback(
                    action_id, custom_tools_path, file_name, path
                )
            ),
            None,
        )

    @classmethod
    def __executed_action_in_traceback(
        cls, action_id, custom_tools_path, file_name, path
    ):
        action_name = action_id.split(".")[-1]
        action_path = os.path.join(custom_tools_path, *action_id.split("."))
        return file_name == action_name or path == action_path

    @classmethod
    def __walk_tb(cls, tb):
        def extract_metadata(_t):
            try:
                p = Path(traceback.extract_tb(_t)[0].filename)
                return p.parent, p.stem
            except IndexError:
                return None

        while tb is not None:
            path, file_name = extract_metadata(tb)
            yield tb, path.__str__(), file_name
            tb = tb.tb_next


class ActionImportError(WithTracebackError):
    def __init__(self, action_id, custom_tools_path) -> None:
        super().__init__(action_id, custom_tools_path)
        self.error = _("error.support.execute.errors.python_dependency_error")


class ActionExecuteError(WithTracebackError):
    def __init__(self, action_id, custom_tools_path) -> None:
        super().__init__(action_id, custom_tools_path)
        self.error = _("error.support.execute.errors.execute_tool_failed").format(
            action=action_id
        )



import sys

from hexagon.runtime.plugins import collect_plugins
from hexagon.support.output.printer import log
from hexagon.support.storage import (
    HexagonStorageKeys,
    store_user_data,
)


def main():
    try:
        from hexagon.runtime.singletons import cli, tools, envs, options, configuration
        from hexagon.runtime.execute.tool import select_and_execute_tool
        from hexagon.runtime.update.cli import check_for_cli_updates
        from hexagon.runtime.update.hexagon import check_for_hexagon_updates
        from hexagon.runtime.help import print_help
        from hexagon.support.hooks import HexagonHooks
        from hexagon.runtime.parse_args import parse_cli_args
        from hexagon.support.tracer import init_tracer
        from hexagon.runtime.version import print_version

        log.load_theme(options.theme)

        args = parse_cli_args()

        if args.show_help:
            return print_help(cli, tools, envs)

        if args.show_version:
            return print_version(cli, tools, envs)

        collect_plugins(configuration.project_path, cli.plugins)
        tracer = init_tracer(args)

        HexagonHooks.start.run()
        log.start(f"[bold]{cli.name}")
        log.gap()

        check_for_hexagon_updates()

        if cli.name == "Hexagon":
            log.info(
                _("msg.main.first_time_intro"), _("msg.main.first_time_tool"), gap_end=1
            )
        else:
            check_for_cli_updates()

        result = select_and_execute_tool(tools, envs, args)

        log.gap()

        if result:
            for item in result:
                log.info(item)

        log.finish()

        tracer.print_run_again(cli.command, log)
        store_user_data(
            HexagonStorageKeys.last_command.value, f"{cli.command} {tracer.trace()}"
        )

        HexagonHooks.end.run()
        return None
    except KeyboardInterrupt:
        sys.exit(1)
    except Exception as e:
        from hexagon.domain.hexagon_error import HexagonError

        if isinstance(e, HexagonError):
            e.print_error(log)
            sys.exit(e.exit_status)
        else:
            raise e


if __name__ == "__main__":
    main()



import sys

from hexagon.runtime.singletons import cli, options
from hexagon.runtime.update.shared import (
    already_checked_for_updates,
    check_cli_version,
    perform_cli_update,
)
from hexagon.runtime.update.silent_fail import silent_fail
from hexagon.support.input.prompt import prompt
from hexagon.support.output.printer import log


@silent_fail
def check_for_cli_updates():
    if options.cli_update_disabled:
        return
    if already_checked_for_updates():
        return

    with log.status(_("msg.support.update.cli.checking_for_cli_updates")) as status:
        needs_update, current_branch = check_cli_version()

        if not needs_update:
            return

        status.update(
            _("msg.support.update.cli.checking_for_cli_updates_on_branch").format(
                branch=current_branch
            )
        )

        log.info(
            _("msg.support.update.cli.new_version_available").format(cli_name=cli.name)
        )

        if not prompt.confirm(
            _("action.support.update.cli.confirm_update"), default=True
        ):
            return

        with log.status(_("msg.support.update.cli.updating")):
            perform_cli_update()

        log.info(_("msg.support.update.cli.updated"))
        log.finish()
        sys.exit(0)



import os
import sys

from hexagon.runtime.singletons import options
from hexagon.runtime.update.shared import (
    already_checked_for_updates,
    check_hexagon_version,
    get_hexagon_changelog,
    perform_hexagon_update,
)
from hexagon.runtime.update.silent_fail import silent_fail
from hexagon.support.input.prompt import prompt
from hexagon.support.output.printer import log
from hexagon.support.storage import HEXAGON_STORAGE_APP


@silent_fail
def check_for_hexagon_updates():
    if options.update_disabled:
        return
    if already_checked_for_updates(HEXAGON_STORAGE_APP):
        return

    with log.status(_("msg.support.update.hexagon.checking_new_versions")):
        current_version, latest_version = check_hexagon_version()

        if current_version >= latest_version:
            return

        log.info(
            _("msg.support.update.hexagon.new_version_available").format(
                latest_version=latest_version
            )
        )

        if bool(os.getenv("HEXAGON_UPDATE_SHOW_CHANGELOG", "1")):
            with log.status(_("msg.support.update.hexagon.fetching_changelog")):
                changelog = get_hexagon_changelog(current_version)

            log.example(changelog, syntax="md")

        if not prompt.confirm(
            _("action.support.update.hexagon.confirm_update"), default=True
        ):
            return

        with log.status(_("msg.support.update.hexagon.updating")):
            perform_hexagon_update()

        log.info(_("msg.support.update.hexagon.updated"))
        log.finish()
        sys.exit(1)



import sys

from hexagon.runtime.update.shared import (
    check_hexagon_version,
    get_hexagon_changelog,
    perform_hexagon_update,
)
from hexagon.support.input.prompt import prompt
from hexagon.support.output.printer import log


def main(tool, env, env_args, cli_args):
    """
    Manually check for and install hexagon updates.

    This bypasses time throttle and always checks PyPI for latest version.
    Shows changelog and prompts for confirmation before updating.
    """
    with log.status(_("msg.support.update.hexagon.checking_new_versions")):
        current_version, latest_version = check_hexagon_version()

    if current_version >= latest_version:
        log.info(
            _("msg.actions.internal.update_hexagon.already_latest").format(
                version=current_version
            )
        )
        return

    log.info(
        _("msg.support.update.hexagon.new_version_available").format(
            latest_version=latest_version
        )
    )

    with log.status(_("msg.support.update.hexagon.fetching_changelog")):
        changelog = get_hexagon_changelog(current_version)

    log.example(changelog, syntax="md")

    if not prompt.confirm(
        _("action.support.update.hexagon.confirm_update"), default=True
    ):
        log.info(_("msg.actions.internal.update_hexagon.cancelled"))
        return

    with log.status(_("msg.support.update.hexagon.updating")):
        perform_hexagon_update()

    log.info(_("msg.support.update.hexagon.updated"))
    log.finish()
    sys.exit(1)



import json
import sys
from urllib.request import Request, urlopen

from packaging.version import parse as parse_version

from hexagon.runtime.update import REPO_NAME


def local(override=None):
    return parse_version(override or _local_version())


def _local_version():
    if sys.version_info >= (3, 8):
        import importlib.metadata as importlib_metadata

        return importlib_metadata.version("hexagon")
    else:
        import pkg_resources

        return pkg_resources.require("hexagon")[0].version


def latest(override: str = None):
    latest_release_request = Request(
        override or f"https://pypi.org/pypi/{REPO_NAME}/json"
    )
    latest_pypi_release = json.load(urlopen(latest_release_request))
    return parse_version(latest_pypi_release["info"]["version"])



import datetime
from typing import Optional, Type

from pydantic import ValidationError
from pydantic.types import DirectoryPath
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    InitSettingsSource,
)

from hexagon.runtime.yaml import YamlValidationError


def _save_settings_to_source(options):
    from hexagon.support.storage import (
        HEXAGON_STORAGE_APP,
        HexagonStorageKeys,
        store_user_data,
    )

    options_dict = options.dict()
    options_dict["update_time_between_checks"] = (
        options.update_time_between_checks.total_seconds()
    )
    store_user_data(
        HexagonStorageKeys.options.value, options_dict, app=HEXAGON_STORAGE_APP
    )


class UserDataSettingsSource(InitSettingsSource):
    def __init__(self, settings_cls: type[BaseSettings]):
        from hexagon.support.storage import (
            HEXAGON_STORAGE_APP,
            HexagonStorageKeys,
            load_user_data,
        )

        local_options = (
            load_user_data(HexagonStorageKeys.options.value, app=HEXAGON_STORAGE_APP)
            or {}
        )

        super().__init__(settings_cls, local_options)


class KeymapOptions(BaseSettings):
    create_dir: str = "c-p"


class Options(BaseSettings):
    model_config = SettingsConfigDict(env_prefix="HEXAGON_")

    theme: Optional[str] = "default"
    update_time_between_checks: Optional[datetime.timedelta] = datetime.timedelta(
        days=1
    )
    send_telemetry: Optional[bool] = None
    disable_dependency_scan: Optional[bool] = False
    update_disabled: Optional[bool] = False
    cli_update_disabled: Optional[bool] = False
    config_storage_path: Optional[DirectoryPath] = None
    hints_disabled: Optional[bool] = False
    keymap: KeymapOptions = KeymapOptions()
    cwd_tools_disabled: Optional[bool] = False

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ):
        return (
            init_settings,
            env_settings,
            UserDataSettingsSource(settings_cls),
            file_secret_settings,
        )


def get_options(init_settings: dict) -> Options:
    try:
        return Options(**init_settings)
    except ValidationError as errors:
        raise YamlValidationError(errors)


def update_options(opt: Options) -> Options:
    _save_settings_to_source(opt)
    return opt.copy()



from typing import List


class ChangelogEntry:
    type: str
    message: str

    def __init__(self, type: str, message: str):
        self.type = type
        self.message = message

    def __str__(self):
        return "{type}: {message}".format(type=self.type, message=self.message)

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return (self.type, self.message) == (other.type, other.message)


class ChangelogVersionEntry:
    version: str
    entries: List[ChangelogEntry]

    def __init__(self, version: str):
        self.version = version
        self.entries = []

    def __str__(self):
        return "{version}\n{entries}".format(
            version=self.version, entries="\n".join([f"\t{e}" for e in self.entries])
        )

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return (self.version, self.entries) == (other.version, other.entries)



from typing import Any, Dict

from hexagon.domain.env import Env
from hexagon.domain.tool import Tool


class ToolExecutionParameters:
    def __init__(
        self, tool: Tool, parameters: Any, env: Env, arguments: Dict[str, Any]
    ) -> None:
        self.tool = tool
        self.parameters = parameters
        self.env = env
        self.arguments = arguments

    tool: Tool
    parameters: Any
    env: Env
    arguments: Dict[str, Any]
    custom_tools_path: str


class ToolExecutionData:
    def __init__(self, tool: Tool, tool_env_args: Any, duration: float) -> None:
        self.tool = tool
        self.tool_env_args = tool_env_args
        self.duration = duration

    tool: Tool
    tool_env_args: Any
    duration: float



import re
from io import StringIO

from markdown import Markdown
from packaging.version import parse as parse_version, Version

from hexagon.runtime.update.changelog.fetch import ChangelogFile

CHANGELOG_MAX_EMPTY_LINES = 10


def parse_changelog(current_hexagon_version: Version, changelog_file: ChangelogFile):
    """
    Parse the changelog file and return a simplified view of it.

    The changelog file is expected to be in the following format:
    ```
    ## v0.7.0
    ### Feature
    * feat(task): added something new ()
    * feat(users): added email support ()
    * feat(users): added legs support ()
    ```

    The function will return the following view:
    ```
    ## v0.7.0
    ### tasks
    feat: added something new
    ### users
    feat: added email support
    feat: added legs support
    ```

    """
    changelog = {}
    current_version = None
    for line in changelog_file.readlines():
        if line.startswith("## v"):
            current_version = parse_version(line.split(" ")[1])
            if current_version == current_hexagon_version:
                break
            changelog[current_version] = {}
        else:
            match = re.match(r"^\* (\w+)\((.*)\):\s(.*)\s(\(.*\))", line)
            if match:
                if match.group(2) in changelog[current_version]:
                    changelog[current_version][match.group(2)].append(
                        f"{match.group(1)}: {match.group(3)}"
                    )
                else:
                    changelog[current_version][match.group(2)] = [
                        f"{match.group(1)}: {match.group(3)}"
                    ]

    return "\n".join(
        [
            "\n".join(
                [f"## v{version}"]
                + [
                    "\n".join([f"### {group}"] + entries)
                    for group, entries in entries.items()
                ]
            )
            for version, entries in changelog.items()
        ]
    )


def _unmark_element(element, stream=None):
    if stream is None:
        stream = StringIO()
    if element.text:
        stream.write(element.text)
    for sub in element:
        _unmark_element(sub, stream)
    if element.tail:
        stream.write(element.tail)
    return stream.getvalue()


Markdown.output_formats["plain"] = _unmark_element
# patching Markdown
# noinspection PyTypeChecker
__md = Markdown(output_format="plain")
__md.stripTopLevelTags = False


def _unmark(text):
    return __md.convert(text)



import os

from hexagon.runtime.dependencies.node import scan_and_install_node_dependencies
from hexagon.runtime.dependencies.python import scan_and_install_python_dependencies
from hexagon.runtime.singletons import options

HEXAGON_DEPENDENCY_UPDATER_MOCK_ENABLED_ENVIRONMENT_VARIABLE = (
    "HEXAGON_DEPENDENCY_UPDATER_MOCK_ENABLED"
)


def scan_and_install_dependencies(path: str = None):
    if path is None:
        return

    if options.disable_dependency_scan:
        return

    mocked = (
        HEXAGON_DEPENDENCY_UPDATER_MOCK_ENABLED_ENVIRONMENT_VARIABLE in os.environ
        and os.environ[HEXAGON_DEPENDENCY_UPDATER_MOCK_ENABLED_ENVIRONMENT_VARIABLE]
        == "1"
    )

    scan_and_install_python_dependencies(path, mocked)

    scan_and_install_node_dependencies(path, mocked)



import abc
from typing import Generic, Union, Callable, TypeVar, get_args, Any, get_origin

from pydantic import GetCoreSchemaHandler
from pydantic_core import core_schema
from pydantic_core.core_schema import ValidatorFunctionWrapHandler

from hexagon.support.input.args.field_reference import FieldReference

ARGUMENT_KEY_PREFIX = "-"

T = TypeVar("T")


class HexagonArg(Generic[T]):
    __model__ = None
    __field__ = None

    def __init__(self, value: T):
        self.__value__ = value

    @property
    def value(self):
        if self.__model__:
            self.__model__.trace(self.__field__)
        return self.__value__

    def prompt(
        self,
        skip_trace: Union[bool, Callable] = False,
        **kwargs,
    ):
        """
        Prompt the user for a value for this argument.

        :param skip_trace: true if this prompt should not be traced by hexagon
        :param kwargs: any extra argument supported by hexagon and InquirePy
        :return:
        """
        if not self.__model__:
            raise ValueError(
                "Cannot prompt for a value when model reference is not initialized. "
                "Probably _init_refs was not called."
            )
        return self.__model__.prompt(self.__field__, skip_trace, **kwargs)

    def _init_refs(self, model: Any, field: FieldReference):
        self.__field__ = field
        self.__model__ = model

    @staticmethod
    @abc.abstractmethod
    def cli_repr(field: FieldReference):
        return

    def __str__(self, **kwargs):
        return str(self.__value__)

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: GetCoreSchemaHandler
    ) -> core_schema.CoreSchema:
        origin = get_origin(source_type)
        if origin is None:  # used as `x: Owner` without params
            origin = source_type
            item_tp = Any
        else:
            item_tp = get_args(source_type)[0]
        # both calling handler(...) and handler.generate_schema(...)
        # would work, but prefer the latter for conceptual and consistency reasons
        item_schema = handler.generate_schema(item_tp)

        def val_item(v: Any, handler: ValidatorFunctionWrapHandler) -> HexagonArg[Any]:
            if not isinstance(v, cls):
                v = cls(handler(v))
            return v

        python_schema = core_schema.chain_schema(
            # `chain_schema` means do the following steps in order:
            [
                # # Ensure the value is an instance of HexagonArg
                # core_schema.is_instance_schema(cls),
                # Use the item_schema to validate `items`
                core_schema.no_info_wrap_validator_function(val_item, item_schema),
            ]
        )

        return core_schema.json_or_python_schema(
            # for JSON accept an object with name and item keys
            json_schema=core_schema.chain_schema(
                [
                    item_schema,
                    # after validating the json data convert it to python
                    core_schema.no_info_before_validator_function(
                        lambda v: cls(v),
                        # note that we re-use the same schema here as below
                        python_schema,
                    ),
                ]
            ),
            python_schema=python_schema,
        )


class PositionalArg(HexagonArg[T]):
    @staticmethod
    def cli_repr(field: FieldReference):
        return field.name, None


class OptionalArg(HexagonArg[T]):
    @staticmethod
    def cli_repr(field: FieldReference):
        def initials():
            return "".join([w[0] for w in field.name.split("_")])

        return (
            name_key(field.name.replace("_", "-")),
            alias_key(
                field.info.alias.replace("_", "-")
                if field.info.alias and field.info.alias != field.name
                else initials()
            ),
        )


def name_key(name: str):
    return f"{ARGUMENT_KEY_PREFIX * 2}{name}"


def alias_key(alias: str):
    return f"{ARGUMENT_KEY_PREFIX}{alias}"



import subprocess
import sys

from hexagon.runtime.singletons import configuration


def output_from_command_in_cli_project_path(command: str) -> str:
    return subprocess.check_output(
        command,
        shell=True,
        cwd=configuration.project_path,
        text=True,
        stderr=subprocess.DEVNULL,
    )


def execute_command_in_cli_project_path(
    command: str, show_stdout: bool = False
) -> None:
    assert (
        subprocess.check_call(
            command,
            shell=True,
            cwd=configuration.project_path,
            stdout=sys.stdout if show_stdout else subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        )
        == 0
    )



from inspect import isclass
from typing import Optional, Dict, Union, List, Callable, get_origin

from pydantic import BaseModel

from hexagon.support.input.args import HexagonArg
from hexagon.support.input.args.field_reference import FieldReference
from hexagon.support.input.args.hexagon_args import PositionalArg, OptionalArg


class ToolArgs(BaseModel):
    model_config = {
        # pydantic config
        "validate_assignment": True,
        # hexagon config
        "trace_on_access": True,
        "trace_on_prompt": True,
        "prompt_on_access": False,
    }
    __tracer__ = None
    __prompt__ = None
    __fields_traced__ = set()
    __fields_initialized__ = set()
    __track_fields_set__ = False

    show_help: bool = False
    extra_args: Optional[Dict[str, Union[list, bool, int, str]]] = None
    raw_extra_args: Optional[List[str]] = None

    def __init__(self, **data):
        for _k, v in self.model_fields.items():
            if (
                isclass(get_origin(v.annotation))
                and issubclass(get_origin(v.annotation), HexagonArg)
            ) and not isinstance(v.default, HexagonArg):
                v.default = (
                    PositionalArg(v.default)
                    if v.annotation == PositionalArg
                    else OptionalArg(v.default)
                )
        super().__init__(**data)
        self.__fields_initialized__ = set(self.model_fields_set)
        # for each self.model_field that is not a HexagonArg, convert it to a HexagonArg
        for k, v in self.model_fields.items():
            if k in ["show_help", "extra_args", "raw_extra_args"]:
                continue
            value_ = self.__getattribute__(k, just_get=True)
            if not isinstance(value_, HexagonArg):
                self.__setattr__(
                    k,
                    (
                        PositionalArg(value_)
                        if get_origin(v.annotation) == PositionalArg
                        else OptionalArg(value_)
                    ),
                )
        self.__track_fields_set__ = True

    def __getattribute__(self, item, just_get=False):
        if just_get:
            return super().__getattribute__(item)
        if item == "model_fields":
            return super().__getattribute__(item)
        if item in self.model_fields:
            field = self.model_fields[item]
            value_ = self.__getattribute__(item, just_get=True)

            if (
                hasattr(value_, "__value__")
                and value_.__value__ is None
                and self.model_config["prompt_on_access"]
            ):
                self.prompt(FieldReference(item, field))
                return self.__getattribute__(item, just_get=True)

        return super().__getattribute__(item)

    def __setattr__(self, key, value):
        super().__setattr__(key, value)
        arg = self.__getattribute__(key, just_get=True)
        if isinstance(arg, HexagonArg):
            # noinspection PyProtectedMember
            arg._init_refs(self, FieldReference(key, self.model_fields.get(key)))
        if self.__track_fields_set__:
            self.__fields_initialized__.add(key)

    def trace(self, field: Union[FieldReference, str], retrace=False):
        if not self.__tracer__:
            raise ValueError("Tracer not initialized. Did _with_tracer() get called?")

        model_field = (
            field
            if isinstance(field, FieldReference)
            else self.model_fields.get(field.name)
        )
        if not model_field:
            raise ValueError(
                f"field [{field}] not found, must be a field name or a FieldReference instance"
            )

        if not self.model_config["trace_on_access"]:
            return

        value_ = self.__getattribute__(model_field.name, just_get=True)

        if (
            retrace or model_field.name not in self.__fields_traced__
        ) and model_field.name in self.__fields_initialized__:
            if get_origin(model_field.info.annotation) == PositionalArg:
                self.__tracer__.tracing(f"arg_{model_field.name}", value_.__value__)
            elif get_origin(model_field.info.annotation) == OptionalArg:
                n, a = OptionalArg.cli_repr(model_field)
                self.__tracer__.tracing(
                    f"arg_{model_field.name}", value_.__value__, key=n, key_alias=a
                )
            self.__fields_traced__.add(model_field.name)

    def prompt(
        self,
        field: Union[FieldReference, str],
        skip_trace: Union[bool, Callable] = False,
        **kwargs,
    ):
        if not self.__prompt__:
            raise ValueError("prompt not initialized. Did _with_prompt() get called?")

        model_field = (
            field
            if isinstance(field, FieldReference)
            else self.model_fields.get(field.name)
        )
        if not model_field:
            raise ValueError(
                f"field [{field}] not found, must be a field name or a FieldReference instance"
            )

        value_ = self.__prompt__.query_field(
            model_field, model_class=self.__class__, **kwargs
        )

        self.__setattr__(model_field.name, value_)
        getattribute__ = self.__getattribute__(model_field.name)

        if callable(skip_trace):
            skip_trace = skip_trace(value_)

        if not skip_trace and self.model_config["trace_on_prompt"]:
            self.trace(model_field, retrace=not skip_trace)

        return getattribute__.__value__

    def _with_tracer(self, tracer):
        self.__tracer__ = tracer
        for k, field in self.model_fields.items():
            arg = self.__getattribute__(k, just_get=True)
            if isinstance(arg, HexagonArg):
                # noinspection PyProtectedMember
                arg._init_refs(self, FieldReference(k, field))
        return self

    def _with_prompt(self, prompt):
        self.__prompt__ = prompt
        for k, field in self.model_fields.items():
            arg = self.__getattribute__(k, just_get=True)
            if isinstance(arg, HexagonArg):
                # noinspection PyProtectedMember
                arg._init_refs(self, FieldReference(k, field))
        return self



import re
from copy import copy
from pathlib import Path
from typing import Any, Dict, Literal, Callable, List

from InquirerPy import inquirer
from InquirerPy.base import Choice
from InquirerPy.utils import run_in_terminal
from prompt_toolkit.buffer import ValidationState
from prompt_toolkit.completion import FuzzyWordCompleter
from prompt_toolkit.document import Document
from prompt_toolkit.validation import ValidationError, Validator
from pydantic import ValidationError as PydanticValidationError, BeforeValidator
from pydantic_core import SchemaValidator

from hexagon.domain.hexagon_error import ListHexagonError
from hexagon.runtime.singletons import options
from hexagon.support.input.args import HexagonArg
from hexagon.support.input.args.field_reference import FieldReference
from hexagon.support.input.prompt.for_all_methods import for_all_methods
from hexagon.support.input.prompt.hints import HintsBuilder
from hexagon.support.input.prompt.inquiry_type import (
    determine_expected_inquiry,
    InquiryType,
)
from hexagon.support.input.types import path_validator
from hexagon.support.output.printer import log
from hexagon.typing import field_type_information, TypeInformation


class HexagonArgumentSetupError(ListHexagonError):
    def __init__(self, argument: str, prop: str):
        super().__init__(
            [
                _("error.support.input.prompt.prompt.invalid_argument_setup").format(
                    argument=argument, property=prop
                ),
            ]
        )


class PromptValidator(Validator):
    def __init__(self, validators, cls):
        self.validators = validators
        self.cls = cls

    def validate(self, document: Document) -> None:
        try:
            for validator in self.validators.values():
                validator.func(self.cls, document.text)
        except PydanticValidationError as e:
            raise ValidationError(
                message=" / ".join([x["msg"] for x in e.errors()]),
                cursor_position=len(document.text),
            )
        except ValueError as e:
            raise ValidationError(message=e.args[0], cursor_position=len(document.text))


def default_validator(field_reference: FieldReference, mapper=lambda x: x):
    def __recurse_schema_until_model_fields_type(schema):
        if schema.get("type") == "model-fields":
            return schema["fields"]
        return __recurse_schema_until_model_fields_type(schema["schema"])

    def func(cls, value):
        schema__ = cls.__pydantic_core_schema__
        target_field_schema: Dict = __recurse_schema_until_model_fields_type(schema__)[
            field_reference.name
        ]["schema"]
        validator = SchemaValidator(
            {
                "type": "definitions",
                "schema": target_field_schema,
                "definitions": schema__.get("definitions"),
            }
            if "definitions" in schema__
            else target_field_schema
        )
        try:
            return validator.validate_python(mapper(value))
        except ValidationError as e:
            raise PydanticValidationError([e], model=cls)

    return func


def list_mapper(v):
    return v.strip().split("\n")


def set_default(invocation_extras, field_reference: FieldReference):
    extras = field_reference.info.json_schema_extra or {}
    if extras.get("prompt_default"):
        return {"default": extras["prompt_default"]}
    if "default" in invocation_extras:
        return {"default": invocation_extras["default"]}
    elif field_reference.info.default:
        default = (
            field_reference.info.default.value
            if isinstance(field_reference.info.default, HexagonArg)
            else field_reference.info.default
        )
        if default is not None:
            return {"default": default}

    return {}


@for_all_methods(log.status_aware, exclude=["query_field"])
class Prompt:
    def query_field(self, field_reference: FieldReference, model_class, **kwargs):
        # TODO: this method is a mess, refactor it
        declaration_extras = copy(field_reference.info.json_schema_extra or {})
        invocation_extras = copy(kwargs)
        inquiry_args = {
            "message": declaration_extras.get("prompt_message", None)
            or f"Enter {field_reference.name}:",
            "instruction": declaration_extras.get("prompt_instruction", None),
        }
        inquiry_args.update(set_default(invocation_extras, field_reference))

        field_type, iterable, of_enum = field_type_information(field_reference.info)

        extras = {
            # **inquiry_args,
            **declaration_extras,
            **invocation_extras,
        }
        inquiry_type = determine_expected_inquiry(
            iterable=iterable,
            of_enum=of_enum,
            field_type=field_type,
            extras=extras,
        )

        setups = {
            InquiryType.ENUM_LIST: setup_enum_list,
            InquiryType.ENUM_LIST_SEARCHABLE: setup_enum_list,
            InquiryType.STRING_LIST: setup_string_list,
            InquiryType.STRING_LIST_SEARCHABLE: setup_searchable_list,
            InquiryType.ENUM: setup_enum,
            InquiryType.ENUM_SEARCHABLE: setup_enum,
            InquiryType.STRING_SEARCHABLE: setup_searchable_list,
            InquiryType.PATH: setup_path,
            InquiryType.PATH_SEARCHABLE: setup_path,
            InquiryType.INT: setup_int,
            InquiryType.FLOAT: setup_float,
            InquiryType.SECRET: setup_secret,
            InquiryType.BOOLEAN: setup_bool,
        }

        inq, mapper = setups.get(inquiry_type, setup_string)(
            self,
            inquiry_args=inquiry_args,
            inquiry_type=inquiry_type,
            extras=extras,
            field_type=field_type,
        )

        inquiry_args["validate"] = PromptValidator(
            {
                "default": BeforeValidator(
                    default_validator(field_reference, mapper=mapper)
                )
            },
            model_class,
        )

        # FIXME: this is working by chance, it's not the best way to do it
        if "searchable" in invocation_extras:
            del invocation_extras["searchable"]

        inquiry_args.update(**invocation_extras)
        try:
            return inq(**inquiry_args)
        except TypeError as e:
            if "__init__() got an unexpected keyword argument" in e.args[0]:
                prop = re.search(r".*\s('\w+')$", e.args[0]).group(1).strip()
                err = HexagonArgumentSetupError(field_reference.name, prop)
            else:
                raise e

        if err:
            raise err
        return None

    @staticmethod
    def text(**kwargs):
        if not options.hints_disabled:
            kwargs["long_instruction"] = (
                HintsBuilder().with_enter_cancel_skip().with_autocomplete().build()
            )
        return inquirer.text(**kwargs).execute()

    @staticmethod
    def select(**kwargs):
        if not options.hints_disabled:
            kwargs["long_instruction"] = (
                HintsBuilder().with_enter_cancel_skip().with_vertical_movement().build()
            )
        return inquirer.select(**kwargs).execute()

    @staticmethod
    def checkbox(**kwargs):
        if not options.hints_disabled:
            kwargs["long_instruction"] = (
                HintsBuilder()
                .with_enter_cancel_skip()
                .with_vertical_movement()
                .with_select_toggles()
                .build()
            )
        return inquirer.checkbox(**kwargs).execute()

    @staticmethod
    def confirm(*args, **kwargs):
        if not options.hints_disabled:
            # not adding confirm_letter and reject_letter as hints because it seems redundant
            kwargs["long_instruction"] = HintsBuilder().with_enter_cancel_skip().build()
        if "validate" in kwargs:
            del kwargs["validate"]  # validate is not supported by confirm
        return inquirer.confirm(*args, **kwargs).execute()

    @staticmethod
    def fuzzy(**kwargs):
        kwargs["keybindings"] = {
            "toggle-exact": [
                {"key": "c-f"}
            ]  # toggle string matching algorithm between fuzzy or exact
        }
        if not options.hints_disabled:
            builder = (
                HintsBuilder()
                .with_enter_cancel_skip()
                .with_vertical_movement()
                .with_fuzzy_toggle()
            )
            if "multiselect" in kwargs:
                builder.with_select_toggles()
            kwargs["long_instruction"] = builder.build()
        return inquirer.fuzzy(border=log.use_borders(), **kwargs).execute()

    @staticmethod
    def path(**kwargs):
        is_dir = kwargs.get("only_directories", False)
        if not options.hints_disabled:
            hints = HintsBuilder().with_enter_cancel_skip().with_autocomplete()
            if is_dir:
                hints.with_path_support()
            kwargs["long_instruction"] = hints.build()
        filepath_prompt = inquirer.filepath(**kwargs)

        if is_dir:
            # noinspection PyProtectedMember
            @filepath_prompt.register_kb(options.keymap.create_dir)
            def _create_dir(__):
                path = path_validator(filepath_prompt._session.default_buffer.text)
                if not path.exists():
                    try:
                        path.mkdir(parents=True)
                        run_in_terminal(
                            lambda: log.info(
                                _("msg.support.prompt.prompt.directory_created").format(
                                    path=path.absolute()
                                )
                            )
                        )
                        filepath_prompt._session.default_buffer.validation_state = (
                            ValidationState.INVALID
                        )
                        filepath_prompt._session.default_buffer.validation_error = None
                    except Exception as e:
                        filepath_prompt._set_error(str(e))

        return filepath_prompt.execute()

    @staticmethod
    def number(**kwargs):
        if not options.hints_disabled:
            builder = HintsBuilder().with_enter_cancel_skip().with_number_controls()
            if "float_allowed" in kwargs:
                builder.with_floating_point()
            kwargs["long_instruction"] = builder.build()
        return inquirer.number(**kwargs).execute()

    @staticmethod
    def secret(**kwargs):
        if not options.hints_disabled:
            kwargs["long_instruction"] = HintsBuilder().with_enter_cancel_skip().build()
        return inquirer.secret(**kwargs).execute()


def setup_enum_list(
    self: Prompt,
    inquiry_type: Literal[InquiryType.ENUM_LIST, InquiryType.ENUM_LIST_SEARCHABLE],
    inquiry_args: Dict[str, Any],
    field_type: TypeInformation,
    **_,
) -> (Callable, Callable):
    inquiry_args["choices"] = [
        Choice(
            name=x.name,
            value=x,
            enabled=("default" in inquiry_args and x in inquiry_args["default"]),
        )
        for x in field_type.base_type.__args__[0]
    ]
    inquiry_args["default"] = ""
    if inquiry_type == InquiryType.ENUM_LIST_SEARCHABLE:
        inquiry_args["multiselect"] = True
        return self.fuzzy, lambda x: x
    else:
        return self.checkbox, lambda x: x


def setup_string_list(
    self: Prompt, inquiry_args: Dict[str, Any], **_
) -> (Callable, Callable):
    mapper = list_mapper
    inquiry_args["filter"] = mapper
    inquiry_args["instruction"] = (
        inquiry_args["instruction"]
        or "(each line represents a value) ESC + Enter to finish input"
    )
    inquiry_args["multiline"] = True
    return self.text, mapper


def setup_enum(
    self: Prompt,
    inquiry_type: Literal[InquiryType.ENUM, InquiryType.ENUM_SEARCHABLE],
    inquiry_args: Dict[str, Any],
    field_type: TypeInformation,
    **_,
) -> (Callable, Callable):
    inquiry_args["choices"] = [
        {"name": x.name, "value": x} for x in field_type.base_type
    ]
    if inquiry_type == InquiryType.ENUM_SEARCHABLE:
        inquiry_args["default"] = (
            inquiry_args["default"].value if "default" in inquiry_args else None
        )
        return self.fuzzy, lambda x: x
    else:
        return self.select, lambda x: x


def setup_searchable_list(
    self: Prompt,
    inquiry_type: Literal[
        InquiryType.STRING_SEARCHABLE, InquiryType.STRING_LIST_SEARCHABLE
    ],
    inquiry_args: Dict[str, Any],
    extras: Dict[str, Any],
    **_,
) -> (Callable, Callable):
    inquiry_args["choices"] = extras["choices"]
    if inquiry_type == InquiryType.STRING_LIST_SEARCHABLE:
        inquiry_args["multiselect"] = True
    return self.fuzzy, lambda x: x


def setup_path(
    self: Prompt,
    inquiry_type: Literal[InquiryType.PATH, InquiryType.PATH_SEARCHABLE],
    inquiry_args: Dict[str, Any],
    extras: Dict[str, Any],
    field_type: TypeInformation,
) -> (Callable, Callable):
    if inquiry_type == InquiryType.PATH_SEARCHABLE:
        choices: List[Dict[str, Any] or Any] = []
        if "glob" in extras:
            choices = choices + [
                f
                for f in sorted(
                    Path(extras.get("cwd", ".")).rglob(extras["glob"]),
                    key=lambda x: x.name,
                )
            ]
        elif "choices" in extras:
            choices = choices + extras["choices"]
        else:
            raise ValueError("searchable path must have either glob or choices defined")
        inquiry_args["choices"] = choices
        return self.fuzzy, lambda x: x
    else:
        inquiry_args["only_directories"] = (
            inquiry_args["only_directories"]
            if "only_directories" in inquiry_args
            else field_type.is_directory_path
        )
        return self.path, lambda x: x


def setup_int(self: Prompt, **_) -> (Callable, Callable):
    return self.number, lambda x: x


def setup_float(
    self: Prompt, inquiry_args: Dict[str, Any], **_
) -> (Callable, Callable):
    inquiry_args["float_allowed"] = True
    return self.number, lambda x: x


def setup_secret(self: Prompt, **_) -> (Callable, Callable):
    return self.secret, lambda x: x


def setup_string(
    self: Prompt, inquiry_args: Dict[str, Any], extras: Dict[str, Any], **_
) -> (Callable, Callable):
    if "prompt_suggestions" in extras:
        inquiry_args["completer"] = FuzzyWordCompleter(extras["prompt_suggestions"])
    return self.text, lambda x: x


def setup_bool(self: Prompt, **_) -> (Callable, Callable):
    return self.confirm, lambda x: x



from typing import Any, get_args, Annotated, get_origin


class TypeInformation:
    def __init__(self, type_info: Any):
        self.type_info = type_info
        self.base_type = (
            get_args(type_info)[0] if get_origin(type_info) is Annotated else type_info
        )

    @property
    def is_bool(self):
        return self.base_type == bool

    @property
    def is_directory_path(self):
        return get_args(self.type_info)[1].path_type == "dir"



from typing import List

from hexagon.domain.cli import Cli
from hexagon.domain.env import Env
from hexagon.domain.tool import Tool
from hexagon.support.output.printer import log

# this updates automatically https://python-semantic-release.readthedocs.io/en/latest/index.html
__version__ = "0.64.1"


def print_version(cli_config: Cli, tools: List[Tool], envs: List[Env]):
    """
    Print either the current installed version of hexagon python package or the version of the cli tool

    :param cli_config:
    :param tools:
    :param envs:
    :return:
    """
    if cli_config.name == "Hexagon":
        log.extra(f"Hexagon: {__version__}")
        return



from typing import Optional, List, Dict, Any

from pydantic import BaseModel


class EntrypointConfig(BaseModel):
    shell: Optional[str] = None
    pre_command: Optional[str] = None
    environ: Dict[str, Any] = {}


class Cli(BaseModel):
    name: str
    command: str
    entrypoint: EntrypointConfig = EntrypointConfig()
    custom_tools_dir: Optional[str] = None
    plugins: List[str] = []
    options: Optional[dict] = None



from typing import List, Tuple

from hexagon.domain.env import Env
from hexagon.domain.hooks.execution import ToolExecutionParameters
from hexagon.domain.tool import (
    FunctionTool,
    GroupTool,
    Tool,
    ToolType,
)
from hexagon.runtime.execute.action import execute_action
from hexagon.runtime.parse_args import parse_cli_args
from hexagon.runtime.wax import search_by_name_or_alias, select_env, select_tool
from hexagon.support.hooks import HexagonHooks
from hexagon.support.input.args import CliArgs
from hexagon.support.tracer import tracer


def select_and_execute_tool(
    tools: List[Tool],
    envs: List[Env],
    cli_args: CliArgs,
    group_ref=0,
) -> List[str]:
    tool = search_by_name_or_alias(tools, cli_args.tool and cli_args.tool.value)
    env = search_by_name_or_alias(envs, cli_args.env and cli_args.env.value)
    # FIXME: validate selected env: if tool has envs defined and env is None -> should fail

    tool = select_tool(tools, tool)
    if tool.traced:
        tracer().tracing(f"tool_{group_ref}", tool.name, value_alias=tool.alias)

    env, tool_env_params = select_env(envs, tool.envs, env)

    if isinstance(tool, GroupTool):
        previous = tools, envs, cli_args
        return _execute_group_tool(
            tool,
            envs,
            cli_args,
            ref=group_ref + 1,
            # If the tool matched the tool argument, disable navigating back
            previous=(
                previous
                if not next(
                    (
                        t
                        for t in tools
                        if cli_args.tool and t.name == cli_args.tool.value
                    ),
                    None,
                )
                else None
            ),
        )

    if isinstance(tool, FunctionTool):
        return tool.function()

    if env:
        tracer().tracing(f"env_{group_ref}", env.name, value_alias=env.alias)

    HexagonHooks.before_tool_executed.run(
        ToolExecutionParameters(
            tool=tool,
            parameters=tool_env_params,
            env=env,
            arguments=cli_args.extra_args,
        )
    )

    return execute_action(tool, tool_env_params, env, cli_args)


GO_BACK_TOOL = Tool(
    name="goback",
    long_name=_("msg.support.execute.tool.go_back_long_name"),
    type=ToolType.function,
    description=_("msg.support.execute.tool.go_back_description"),
    icon=_("icon.global.go_back"),
    traced=False,
)


def _execute_group_tool(
    tool: GroupTool,
    envs: List[Env],
    cli_args: CliArgs,
    ref,
    previous: Tuple = None,
) -> List[str]:
    tools = tool.tools

    if previous:

        def go_back():
            tracer().remove_last()
            select_and_execute_tool(*previous)

        # FunctionTool is not set as a type GroupTool.tools
        # so yaml validations don't show that tool.function is required
        # noinspection PyTypeChecker
        tools = tool.tools + [
            FunctionTool(**GO_BACK_TOOL.dict(), function=go_back),
        ]

    return select_and_execute_tool(
        tools,
        envs,
        parse_cli_args(
            ([cli_args.env.value] if cli_args.env else []) + cli_args.raw_extra_args
        ),
        group_ref=ref,
    )



from enum import Enum
from typing import Any, Dict, Generic, TypeVar


class SelectionType(Enum):
    prompt = "prompt"
    args = "args"


SelectedObject = TypeVar("SelectedObject")


class Selection(Generic[SelectedObject]):
    def __init__(self, selected: SelectedObject, type: SelectionType, **kwargs) -> None:
        self.selected = selected
        self.type = type
        self.additional_data = kwargs

    type: SelectionType
    selected: SelectedObject
    additional_data: Dict[str, Any]



from pathlib import Path
from typing import List, Dict


def declarations_found(path: str, dependency_files: List[str]):
    matches: Dict[Path, List[str]] = {}

    for f in dependency_files:
        for p in Path(path).rglob(f):
            if p.parent not in matches:
                matches[p.parent] = []
            matches[p.parent].append(p.name)

    return list(matches.items())



import os
from dataclasses import dataclass, fields, field


@dataclass
class PromptsTheme:
    questionmark: str = "#e5c07b"
    answer: str = "#61afef"
    input: str = "#98c379"
    question: str = ""
    instruction: str = "#6bb4b4"
    long_instruction: str = "orange"
    pointer: str = "#61afef"
    checkbox: str = "#98c379"
    separator: str = ""
    skipped: str = "#5c6370"
    validator: str = ""
    marker: str = ""
    fuzzy_prompt: str = "#c678dd"
    fuzzy_info: str = "#56b6c2"
    fuzzy_border: str = "#4b5263"
    fuzzy_match: str = "#c678dd"


@dataclass
class LoggingTheme:
    show_colors: bool = True
    result_only: bool = False
    prompt_border: bool = True
    start: str = ""
    border: str = ""
    border_result: str = ""
    process_out: str = ""
    process_in: str = ""
    finish: str = ""
    prompts: PromptsTheme = field(default_factory=PromptsTheme)

    def __getitem__(self, item: str):
        if item.startswith("prompts."):
            return (
                getattr(self.prompts, item.replace("prompts.", ""))
                if self.show_colors
                else ""
            )
        return getattr(self, item)


def load_theme(theme: str):
    __themes = {
        "default": LoggingTheme(
            start="╭╼ ",
            border="│ ",
            border_result="├ ",
            process_out="┆",
            process_in="┆",
            finish="╰╼ ",
        ),
        "no_border": LoggingTheme(
            prompt_border=False,
            start="╭╼ ",
            border="│ ",
            border_result="├ ",
            process_out="┆",
            process_in="┆",
            finish="╰╼ ",
        ),
        "disabled": LoggingTheme(show_colors=False, prompt_border=False),
        "result_only": LoggingTheme(
            result_only=True, show_colors=False, prompt_border=False
        ),
    }

    t = __themes[theme]

    for f in fields(t.prompts):
        n = f"INQUIRERPY_STYLE_{f.name.upper()}"
        os.environ[n] = os.getenv(n, t[f"prompts.{f.name}"])

    return t



from typing import Optional, Union, Any, Callable, List

from pydantic.fields import (
    Field as PydanticField,
)

from hexagon.support.input.args.cli_args import CliArgs
from hexagon.support.input.args.hexagon_args import (
    ARGUMENT_KEY_PREFIX,
    HexagonArg,
    PositionalArg,
    OptionalArg,
)
from hexagon.support.input.args.tool_args import ToolArgs

ARGUMENT_KEY_PREFIX = ARGUMENT_KEY_PREFIX
HexagonArg = HexagonArg
PositionalArg = PositionalArg
OptionalArg = OptionalArg
ToolArgs = ToolArgs
CliArgs = CliArgs


# noinspection PyPep8Naming
def Arg(
    default: Any = None,
    *,
    alias: Optional[str] = None,
    title: Optional[str] = None,
    description: Optional[str] = None,
    prompt_default: Optional[Union[Any, Callable[[Any], Any]]] = None,
    prompt_message: Optional[Union[str, Callable[[Any], str]]] = None,
    prompt_instruction: Optional[str] = None,
    prompt_suggestions: Optional[Union[List[str], Callable[[], List[str]]]] = None,
    searchable: bool = False,
    **kwargs,
):
    """
    Used to provide extra information about an argument, either for the model schema or complex validation.
    Some arguments apply only to number fields (``int``, ``float``, ``Decimal``) and some apply only to ``str``.

    TODO: add support for `validators` kwarg
    """
    kwargs.update(
        [
            (k, v)
            for k, v in {
                "prompt_default": prompt_default,
                "prompt_message": prompt_message,
                "prompt_instruction": prompt_instruction,
                "prompt_suggestions": prompt_suggestions,
                "searchable": searchable,
            }.items()
            if v is not None
        ]
    )
    return PydanticField(
        default,
        alias=alias,
        title=title,
        description=description,
        **kwargs,
    )



import os
from shutil import copytree

from pydantic import DirectoryPath

from hexagon.actions import external
from hexagon.domain.tool import ActionTool, ToolType
from hexagon.runtime.singletons import configuration
from hexagon.support.input.args import ToolArgs, PositionalArg, Arg, OptionalArg
from hexagon.support.output.printer import log


class Args(ToolArgs):
    action: PositionalArg[str] = Arg(
        None, prompt_message=_("action.actions.internal.create_new_tool.choose_action")
    )
    type: OptionalArg[ToolType] = Arg(
        None, prompt_message=_("action.actions.internal.create_new_tool.choose_type")
    )
    name: OptionalArg[str] = Arg(
        None, prompt_message=_("action.actions.internal.create_new_tool.input_name")
    )
    alias: OptionalArg[str] = Arg(
        None, prompt_message=_("action.actions.internal.create_new_tool.input_alias")
    )
    long_name: OptionalArg[str] = Arg(
        None,
        prompt_message=_("action.actions.internal.create_new_tool.input_long_name"),
    )
    description: OptionalArg[str] = Arg(
        None,
        prompt_message=_("action.actions.internal.create_new_tool.input_description"),
    )
    custom_tools_path: OptionalArg[DirectoryPath] = Arg(
        ".",
        prompt_message=_(
            "action.actions.internal.create_new_tool.input_custom_tools_path"
        ),
    )


def main(tool, env, env_args, cli_args):
    create_action = False
    new_tool = ActionTool(
        name="invalid",
        action=cli_args.action.prompt(
            searchable=True,
            choices=external.__all__ + ["new_action"],
            validate=lambda x: x,
        ),
    )

    if new_tool.action == "new_action":
        create_action = True
        new_tool.action = cli_args.action.prompt(
            message=_("action.actions.internal.create_new_tool.input_action"),
            validate=lambda x: x,
        )

    new_tool.type = cli_args.type.prompt(
        choices=[
            {"value": ToolType.web, "name": ToolType.web.value},
            {"value": ToolType.shell, "name": ToolType.shell.value},
        ],
        default=ToolType.web if new_tool.action == "open_link" else ToolType.shell,
    )

    new_tool.name = cli_args.name.prompt(
        validate=lambda x: x,
        default=new_tool.action.replace("_", "-"),
    )

    new_tool.alias = cli_args.alias.prompt(
        default="".join([z[:1] for z in new_tool.name.split("-")]),
        filter=lambda r: r or None,
    )

    new_tool.long_name = cli_args.long_name.prompt(filter=lambda r: r or None)

    new_tool.description = cli_args.description.prompt(filter=lambda r: r or None)

    cli, tools, envs = configuration.refresh()

    if create_action:
        if not configuration.custom_tools_path:
            log.info(_("msg.actions.internal.create_new_tool.custom_tools_dir_not_set"))
            path_ = cli_args.custom_tools_path.prompt()
            configuration.update_custom_tools_path(
                (
                    path_.resolve()
                    if path_.is_absolute()
                    else os.path.join(*path_.parents, path_.name)
                ),
                comment=_(
                    "msg.actions.internal.create_new_tool.input_custom_tools_path_comment"
                ),
            )

        copytree(
            os.path.abspath(
                os.path.join(
                    os.path.dirname(__file__),
                    "..",
                    "..",
                    "actions",
                    "__templates",
                    "custom_tool",
                )
            ),
            os.path.join(configuration.custom_tools_path, new_tool.action),
            dirs_exist_ok=True,
        )

        _replace_variables(
            os.path.join(configuration.custom_tools_path, new_tool.action, "README.md"),
            "{{tool}}",
            new_tool.long_name or new_tool.name,
        )

    configuration.add_tool(new_tool)
    configuration.save()


def _replace_variables(file_name, variable, value):
    with open(file_name, "r") as file:
        file_data = file.read()

    file_data = file_data.replace(variable, value)

    with open(file_name, "w") as file:
        file.write(file_data)



import os
from urllib.request import Request, urlopen

from hexagon.runtime.update.github import add_github_access_token


class ChangelogFile(object):
    def __init__(self, file):
        self.file = file

    def readlines(self):
        raise NotImplementedError("Please Implement this method")


class LocalChangelogFile(ChangelogFile):
    def readlines(self):
        return self.file.readlines()


class RemoteChangelogFile(ChangelogFile):
    def readlines(self):
        charset = self.file.headers.get_content_charset() or "utf-8"
        return [line.decode(charset) for line in self.file.readlines()]


def fetch_changelog(repo_org: str, repo_name: str):
    request = Request(
        f"https://api.github.com/repos/{repo_org}/{repo_name}/contents/CHANGELOG.md"
    )
    request.add_header("Accept", "application/vnd.github.3.raw")
    add_github_access_token(request)
    changelog_file_path = os.getenv("HEXAGON_CHANGELOG_FILE_PATH_TEST_OVERRIDE")
    return (
        LocalChangelogFile(open(changelog_file_path, "r"))
        if changelog_file_path
        else RemoteChangelogFile(urlopen(request))
    )



from typing import Union, List

from hexagon.domain.env import Env
from hexagon.domain.hooks.wax import Selection, SelectionType
from hexagon.domain.tool import Tool, Separator
from hexagon.support.hooks import HexagonHooks
from hexagon.support.input.prompt import prompt


def __classifier(value: Union[Tool, Env]):
    if value.icon:
        return f"{value.icon:2}"

    symbols = {
        "web": "⦾",
        "shell": "ƒ",
        "misc": " ",
        "hexagon": "⬡",
        "group": "≡",
    }
    r = symbols.get(value.type, "")
    return f"{r:2}" if r else ""


def __choices_with_long_name(choices: List[Union[Tool, Env]], classifier=lambda x: ""):
    def build_display(v: Union[Tool, Env]):
        if Separator.name in v.name:
            return "--------------------------------------------------------------------------------"
        else:
            gap = 60 if v.description else 0
            return f"{classifier(v) + (v.long_name or v.name): <{gap}}{v.description or ''}"

    return [{"value": each.name, "name": build_display(each)} for each in choices]


def search_by_name_or_alias(data: List[Union[Tool, Env]], arg: str):
    if arg:
        for k in data:
            if k.name == arg or k.alias == arg:
                return k.name

    return None


def select_env(available_envs: List[Env], tool_envs: dict = None, selected: str = None):
    if not tool_envs:
        return None, None

    if "*" in tool_envs:
        return None, tool_envs["*"]

    (env, prompted) = (
        (selected, False)
        if selected
        else (
            prompt.fuzzy(
                message=_("action.support.wax.select_environment"),
                choices=__choices_with_long_name(
                    [e for e in available_envs if e.name in tool_envs.keys()]
                ),
                validate=lambda x: x and Separator.name not in x,
                invalid_message=_("error.support.wax.invalid_environment"),
            ),
            True,
        )
    )

    return (
        _select_and_register_event(env, available_envs, prompted, target="env"),
        tool_envs[env],
    )


def select_tool(tools: List[Tool], selected: str = None):
    if selected:
        return _select_and_register_event(selected, tools, target="tool")

    name = prompt.fuzzy(
        message=_("action.support.wax.select_tool"),
        choices=__choices_with_long_name(tools, classifier=__classifier),
        validate=lambda x: x and Separator.name not in x,
        invalid_message=_("error.support.wax.invalid_tool"),
    )

    return _select_and_register_event(name, tools, target="tool", prompt=True)


def _select_and_register_event(
    name: str, options: List[Union[Tool, Env]], prompt=False, **kwargs
):
    selected = next((e for e in options if e.name == name), None)
    if selected:
        selection = Selection(
            selected, SelectionType.prompt if prompt else SelectionType.args, **kwargs
        )
        if isinstance(selected, Tool):
            HexagonHooks.tool_selected.run(selection)
        elif isinstance(selected, Env):
            HexagonHooks.env_selected.run(selection)
    return selected



from pathlib import Path
from typing import Any

from pydantic import (
    FilePath as PydanticFilePath,
    DirectoryPath as PydanticDirectoryPath,
    errors,
)

FilePath = PydanticFilePath

DirectoryPath = PydanticDirectoryPath


def path_validator(v: Any) -> Path:
    if isinstance(v, Path):
        return v

    try:
        return Path(v)
    except TypeError:
        raise errors.PathError()



from types import TracebackType
from typing import Optional, Union, Type, Literal

from rich.console import Console
from rich.panel import Panel
from rich.syntax import Syntax
from rich.text import Text

from hexagon.support.output.printer.themes import load_theme, LoggingTheme


class Logger:
    def __init__(
        self, theme: Union[str, LoggingTheme], console: Console = None
    ) -> None:
        self.__console = None
        self.__decorations = None
        self.__live_status = None
        if isinstance(theme, str):
            self.load_theme(theme)
        else:
            self.load_theme(theme, console)

    def start(self, message: str):
        """
        Print a start message to the console.

        Usually used at the beginning of the CLI invocation.
        ie:
        ╭╼ My CLI

        :param message: the message to be printed
        """
        if not self.__decorations.result_only:
            self.__console.print(f"{self.__decorations.start}{message}")

    def gap(self, repeat: int = 1):
        """
        Print a gap to the console.

        ie:
        log.gap(2)
        log.info("some output")

        result:
        │
        │
        │ some output

        :param repeat: the number of times the gap should be printed
        """
        if not self.__decorations.result_only:
            for _ in range(repeat):
                self.__console.print(self.__decorations.border)

    def info(self, *message: str, gap_start: int = 0, gap_end: int = 0):
        """
        Print an informatory message to the console.
        This the most common type of message to be printed.

        ie:
        log.info("some output")

        result:
        │ some output

        :param message: the message to be printed
        :param gap_start: the number of gaps to be printed before the message
        :param gap_end: the number of gaps to be printed after the message
        """
        if not self.__decorations.result_only:
            self.gap(gap_start)
            for msg in message:
                self.__console.print(f"{self.__decorations.border}{msg}")
            self.gap(gap_end)

    def result(self, message: str):
        """
        Print a result message to the console.
        Result messages are usually the output of a command.

        ie:
        log.result("some output")

        result:
        ├ some output

        :param message: the message to be printed
        """
        self.__console.print(f"{self.__decorations.border_result}{message}")

    def example(
        self,
        message: Union[str, Syntax],
        syntax: Optional[str] = None,
        show_line_numbers: bool = False,
        line_range: Optional[tuple] = None,
        decorator_start: Union[str, bool] = True,
        decorator_end: Union[str, bool] = True,
        **kwargs,
    ):
        """
        Print an example to the console.
        Examples are usually code snippets, file contents, shell commands,
        basically anything that should be an example to the user.

        ie:
        log.info("running command")
        log.example("$ docker run -it ubuntu:latest")

        result:
        │ running command
        ┆

        $ docker run -it ubuntu:latest

        ┆

        :param message: the message to be printed
        :param syntax: the syntax highlighting to be used
        :param show_line_numbers: whether to show line numbers
        :param line_range: range of lines to be printed
        :param decorator_start: the decorator to be printed before the message
        :param decorator_end: the decorator to be printed after the message
        :param kwargs: additional arguments to be passed to the Syntax object
        """

        def __use_decorator(param, default):
            return param if isinstance(param, str) else default

        if not self.__decorations.result_only and decorator_start is not False:
            self.__console.print(
                __use_decorator(decorator_start, f"{self.__decorations.process_out}\n")
            )

        if not isinstance(message, Syntax) and syntax:
            message = Syntax(
                message,
                lexer=syntax,
                line_numbers=show_line_numbers,
                line_range=line_range,
                **kwargs,
            )

        self.__console.print(message)

        if not self.__decorations.result_only and decorator_end is not False:
            self.__console.print(
                __use_decorator(decorator_end, f"\n{self.__decorations.process_in}")
            )

    def file(
        self,
        path: str,
        encoding: str = "utf-8",
        syntax: Optional[str] = None,
        show_line_numbers: bool = True,
        line_range: Optional[tuple] = None,
        **kwargs,
    ):
        """
        Pretty print a file from the specified path to the console.
        The syntax highlighting will be inferred from the file extension.

        :param path: Path to the file to be printed
        :param encoding: Encoding of the file
        :param syntax: Syntax to be used for highlighting, defaults to inferred
        :param show_line_numbers: Whether to show line numbers
        :param line_range: Range of lines to be printed
        :param kwargs: Additional arguments to be passed to the Syntax object
        """
        self.__console.print(
            Syntax.from_path(
                path,
                encoding=encoding,
                lexer=syntax,
                line_numbers=show_line_numbers,
                line_range=line_range,
                **kwargs,
            )
        )

    def panel(
        self,
        message,
        title: str = None,
        footer: str = None,
        color: str = None,
        fit: bool = True,
        padding: tuple = None,
        justify: Literal["default", "left", "center", "right", "full"] = None,
        **kwargs,
    ):
        """
        Print a panel to the console.
        A panel is a block of text that is highlighted.

        ie:
        log.panel("No hexagon bins path configured yet.")

        result:
        ╭──────────────────────────────────────────────╮
        │                                              │
        │     No hexagon bins path configured yet.     │
        │                                              │
        ╰──────────────────────────────────────────────╯

        :param message: the message to be printed inside the panel
        :param title: the title of the panel
        :param footer: the footer of the panel
        :param color: the color of the panel borders and text
        :param fit: whether the panel should fit the content or not
        :param padding: the padding of the panel
        :param justify: the justification of the text inside the panel
        :param kwargs: additional arguments to be passed to the Panel object
        """
        color = color or "yellow"
        self.extra(
            Panel(
                Text.from_markup(message, justify=justify, style=color),
                style=color,
                padding=padding or (1, 5),
                title=title,
                subtitle=footer,
                expand=not fit,
                **kwargs,
            )
        )

    def error(self, message: str, err: Optional[Exception] = None):
        """
        Print a red error message to the console.
        Error messages are usually used to display errors to the user.

        ie:
        log.error("an error occurred")

        result:
        [red]an error occurred

        :param message: the message to be printed
        :param err: the exception to be printed, if any
        """
        self.__console.print(f"[red]{message}")
        if err:
            self.__console.print(err)

    def extra(self, *message: Union[str, Panel]):
        """
        Useful for printing additional information to the console.

        ie:
        log.extra("To run this tool again do:")
        log.extra("     hexagon install")

        result:
        To run this tool again do:
             hexagon install

        :param message: the messages to be printed
        """
        if not self.__decorations.result_only:
            for msg in message:
                self.__console.print(msg)

    def finish(self, message: str = None):
        """
        Print a closing message to the console.
        Usually used at the end of the CLI invocation.

        ie:
        log.finish("done")

        result:
        ╰╼ done

        :param message: the message to be printed
        """
        if not self.__decorations.result_only:
            self.__console.print(f"{self.__decorations.finish}{message or ''}")

    def status(self, message: str = None):
        """
        Show a progress status icon to the user.
        Useful for long-running commands.

        ie:
        with log.status("running command"):
            # command execution
            pass

        result:
        ⠧ running command

        :param message: the message to be shown alongside the status icon
        """
        if self.__live_status:
            self.__live_status.update(message)
        else:
            self.__live_status = NestableStatus(self.__console.status(message))
        return self.__live_status

    def load_theme(self, theme: Union[str, LoggingTheme], console: Console = None):
        self.__decorations = (
            theme if isinstance(theme, LoggingTheme) else load_theme(theme)
        )
        self.__console = console or Console(
            color_system="auto" if self.__decorations.show_colors else None
        )

    def use_borders(self):
        return self.__decorations.prompt_border

    def status_aware(self, decorated):
        """
        Decorator that stops the status bar while the decorated function is running.

        :param decorated: the function to be decorated
        :return: the decorated function
        """

        def decorator(cls, *args, **kwargs):
            stopped = False
            if self.__live_status and self.__live_status.active:
                self.__live_status.stop()
                stopped = True

            res = decorated(*args, **kwargs)

            if stopped:
                self.__live_status.start()
            return res

        return decorator


class NestableStatus:
    """
    A wrapper for rich's Status class that allows nesting status calls.

    ie.:
    with log.status("foo"):
        # status should show "foo"
        with log.status("bar"):
            # status should show "bar"
        # status should show "foo"
    # status should be hidden

    """

    def __init__(self, status):
        self.__status = status
        self.__levels = 1

    @property
    def renderable(self):
        return self.__status.renderable

    @property
    def console(self):
        return self.__status.console

    def update(self, *args, **kwargs):
        self.__levels += 1
        self.__status.update(*args, **kwargs)

    def start(self):
        self.__status.start()

    def stop(self) -> None:
        self.__status.stop()

    def __rich__(self):
        return self.__status.__rich__()

    def __enter__(self):
        self.__status.start()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ):
        self.__levels -= 1
        if self.__levels <= 1:
            self.__status.__exit__(exc_type, exc_val, exc_tb)

    @property
    def active(self):
        return self.__levels > 1



from hexagon.domain.hexagon_error import HexagonError


class MultipleHintsNotSupportedError(HexagonError):
    def __init__(self, types: tuple) -> None:
        self.message = f"Multiple hints not supported: {types}"
        super().__init__(self._error_printer)

    def _error_printer(self, logger):
        logger.error(self.message)



def for_all_methods(decorator, exclude=None):
    exclude = exclude or []

    def decorate(cls):
        for attr in cls.__dict__:
            if callable(getattr(cls, attr)) and attr not in exclude:
                setattr(cls, attr, decorator(getattr(cls, attr)))
        return cls

    return decorate



import sys

from hexagon.runtime.singletons import cli
from hexagon.runtime.update.shared import check_cli_version, perform_cli_update
from hexagon.support.input.prompt import prompt
from hexagon.support.output.printer import log


def main(tool, env, env_args, cli_args):
    """
    Manually check for and install CLI updates.

    This bypasses time throttle and always checks git for updates.
    Prompts for confirmation before updating.
    """
    with log.status(_("msg.support.update.cli.checking_for_cli_updates")) as status:
        needs_update, current_branch = check_cli_version()

        if not needs_update:
            if current_branch:
                log.info(
                    _("msg.actions.internal.update_cli.already_latest").format(
                        cli_name=cli.name, branch=current_branch
                    )
                )
            else:
                log.info(
                    _("msg.actions.internal.update_cli.no_git_config").format(
                        cli_name=cli.name
                    )
                )
            return

        status.update(
            _("msg.support.update.cli.checking_for_cli_updates_on_branch").format(
                branch=current_branch
            )
        )

    log.info(
        _("msg.support.update.cli.new_version_available").format(cli_name=cli.name)
    )

    if not prompt.confirm(_("action.support.update.cli.confirm_update"), default=True):
        log.info(_("msg.actions.internal.update_cli.cancelled"))
        return

    with log.status(_("msg.support.update.cli.updating")):
        perform_cli_update()

    log.info(_("msg.support.update.cli.updated"))
    log.finish()
    sys.exit(0)



import subprocess
import sys
from typing import Optional

from hexagon.runtime.dependencies.fs import declarations_found
from hexagon.support.output.printer import log

PACKAGE_JSON_FILE_NAME = "package.json"
PACKAGE_JSON_LOCK_FILE_NAME = "package-lock.json"
YARN_LOCK_FILE_NAME = "yarn.lock"
NODEJS_DECLARATION_FILES = [
    PACKAGE_JSON_FILE_NAME,
    PACKAGE_JSON_LOCK_FILE_NAME,
    YARN_LOCK_FILE_NAME,
]


def scan_and_install_node_dependencies(path: str, mocked=False):
    for directory, files in declarations_found(path, NODEJS_DECLARATION_FILES):
        with log.status(
            _("msg.support.dependencies.installing_dependencies").format(
                runtime="node", path=path
            )
        ):
            command: Optional[str] = None
            if PACKAGE_JSON_FILE_NAME in files:
                if (
                    YARN_LOCK_FILE_NAME in files
                    and PACKAGE_JSON_LOCK_FILE_NAME in files
                ):
                    command = "npm install --only=production"
                elif YARN_LOCK_FILE_NAME in files:
                    command = "yarn --production"
                else:
                    command = "npm install --only=production"

            if mocked:
                print(f"would have ran {command}")
            else:
                subprocess.check_call(
                    command,
                    shell=True,
                    cwd=directory,
                    stdout=sys.stdout,
                    stderr=subprocess.DEVNULL,
                )



from enum import Enum
from typing import Any, Union, get_args, get_origin, Annotated

from pydantic.fields import FieldInfo

from hexagon.support.input.args import HexagonArg
from hexagon.support.input.args.errors import MultipleHintsNotSupportedError
from hexagon.typing.type_information import TypeInformation


def should_support_multiple_args(field: Union[FieldInfo, Any]):
    type_ = _field_type(field) if isinstance(field, FieldInfo) else field
    if hasattr(type_, "__origin__"):
        type_ = type_.__origin__
    return type_ in [list, tuple, set]


def field_type_information(field: FieldInfo):
    type_ = _field_type(field)
    iterable = should_support_multiple_args(field)
    if iterable and hasattr(type_, "__args__") and len(type_.__args__):
        of_enum = _is_enumerable(type_.__args__[0])
    else:
        of_enum = _is_enumerable(type_)

    return TypeInformation(type_), iterable, of_enum


def _field_type(field: FieldInfo):
    p = None
    t = field.annotation
    ars = get_args(t)
    while __needs_unpacking(t):
        p = t
        ars = get_args(t)
        t = ars[0]
    if get_origin(p) is Annotated:
        return p
    if len(ars) > 1 and ars[-1] is not type(None):
        raise MultipleHintsNotSupportedError(ars)
    return t


def __needs_unpacking(t):
    try:
        if (
            get_origin(t) is Union
            or issubclass(get_origin(t), HexagonArg)
            or get_origin(t) is Annotated
        ):
            return True
    except TypeError:
        return False


def _is_enumerable(type_):
    try:
        res = issubclass(type_, Enum)
    except TypeError:
        res = False
    return res



import datetime
import os
import re
import subprocess
import sys
from typing import Optional, Tuple

from packaging.version import Version

from hexagon.runtime.singletons import options
from hexagon.support.storage import (
    HexagonStorageKeys,
    load_user_data,
    store_user_data,
)

LAST_UPDATE_DATE_FORMAT = "%Y%m%d"


def already_checked_for_updates(app: str = None) -> bool:
    last_checked = load_user_data(HexagonStorageKeys.last_update_check.value, app)

    result = False

    if last_checked:
        last_checked_date = datetime.datetime.strptime(
            last_checked, LAST_UPDATE_DATE_FORMAT
        )

        result = (
            last_checked_date + options.update_time_between_checks
            >= datetime.datetime.now()
        )

    if not result:
        store_user_data(
            HexagonStorageKeys.last_update_check.value,
            datetime.date.today().strftime(LAST_UPDATE_DATE_FORMAT),
            app=app,
        )

    return result


def check_hexagon_version() -> Tuple[Version, Version]:
    """
    Check current and latest hexagon versions.

    Returns:
        Tuple of (current_version, latest_version)
    """
    from hexagon.runtime.update import version

    current_version = version.local(
        override=os.getenv("HEXAGON_TEST_LOCAL_VERSION_OVERRIDE", None)
    )
    latest_version = version.latest(
        override=os.getenv("HEXAGON_TEST_LATEST_VERSION_OVERRIDE", None)
    )
    return current_version, latest_version


def perform_hexagon_update() -> None:
    """Execute pip upgrade command for hexagon."""
    subprocess.check_call(
        f"{sys.executable} -m pip install hexagon --upgrade",
        shell=True,
        stdout=subprocess.DEVNULL,
    )


def get_hexagon_changelog(current_version: Version) -> str:
    """
    Fetch and parse changelog from GitHub.

    Args:
        current_version: Current hexagon version

    Returns:
        Formatted changelog string
    """
    from hexagon.runtime.update import REPO_ORG, REPO_NAME
    from hexagon.runtime.update.changelog.fetch import fetch_changelog
    from hexagon.runtime.update.changelog.parse import parse_changelog

    return parse_changelog(current_version, fetch_changelog(REPO_ORG, REPO_NAME))


def check_cli_version() -> Tuple[bool, Optional[str]]:
    """
    Check if CLI needs update.

    Returns:
        Tuple of (needs_update, current_branch)
    """
    from hexagon.runtime.update.cli.command import (
        execute_command_in_cli_project_path,
        output_from_command_in_cli_project_path,
    )
    from hexagon.runtime.update.cli.git import load_cli_git_config

    cli_git_config = load_cli_git_config()
    if not cli_git_config:
        return False, None

    current_branch_status = output_from_command_in_cli_project_path("git status")
    current_branch_match = re.search(r"On branch (.+)", current_branch_status)
    if not current_branch_match:
        return False, None
    current_branch = current_branch_match.group(1)

    execute_command_in_cli_project_path("git remote update")
    branch_status = output_from_command_in_cli_project_path("git status -uno")

    needs_update = "is behind" in branch_status
    return needs_update, current_branch


def perform_cli_update() -> None:
    """Execute git pull and scan dependencies."""
    from hexagon.runtime.dependencies import scan_and_install_dependencies
    from hexagon.runtime.singletons import configuration
    from hexagon.runtime.update.cli.command import execute_command_in_cli_project_path

    execute_command_in_cli_project_path("git pull", show_stdout=True)
    scan_and_install_dependencies(configuration.custom_tools_path)



import glob
import json
from os.path import dirname

import setuptools

# this updates automatically https://python-semantic-release.readthedocs.io/en/latest/index.html
__version__ = "0.64.1"


def __markers(config: dict):
    return f"; {config['markers']}" if "markers" in config else ""


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("Pipfile.lock", "r", encoding="utf-8") as lock:
    json_lock = json.load(lock)
    requires = [
        f"{name}{config['version']}{__markers(config)}"
        for name, config in json_lock["default"].items()
        if "version" in config
    ]

translations = [(dirname(x), [x]) for x in glob.glob("locales/*/LC_MESSAGES/*.mo")]

setuptools.setup(
    name="hexagon",
    version=__version__,
    author="Joaco Campero",
    author_email="juacocampero@gmail.com",
    description="Build your Team's CLI",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/lt-mayonesa/hexagon",
    project_urls={"Bug Tracker": "https://github.com/lt-mayonesa/hexagon/issues"},
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(
        exclude=["tests", "tests.*", "tests_e2e", "tests_e2e.*"]
    ),
    include_package_data=True,
    install_requires=requires,
    python_requires=">=3.9",
    entry_points="""
        [console_scripts]
        hexagon=hexagon.__main__:main
    """,
    data_files=translations,
)



import re

from hexagon.runtime.execute.tool import select_and_execute_tool
from hexagon.runtime.parse_args import parse_cli_args
from hexagon.runtime.singletons import tools, envs
from hexagon.support.output.printer import log
from hexagon.support.storage import (
    HexagonStorageKeys,
    load_user_data,
)


def main(tool, env, env_args, cli_args):
    last_command = load_user_data(HexagonStorageKeys.last_command.value)

    log.info(
        _("msg.actions.internal.replay.last_command").format(last_command=last_command)
    )

    to_list = __command_str_to_list(last_command)
    args = parse_cli_args(to_list[1:])

    return select_and_execute_tool(tools, envs, args)


def __command_str_to_list(last_command: str):
    """
    last_command will be something like:
    hexagon-test python-module john 10 --country="The Netherlands" --likes=orange
    or
    hexagon-test python-module john 10 --country='The Netherlands' --likes=orange
    simulate sys.argv behavior and return something like:
    ['hexagon-test', 'python-module', 'john', '10', '--country=The Netherlands', '--likes=orange']
    """
    split = re.findall(r'(?:"[^"]*"|\'[^\']*\'|[^\s"\'])+', last_command)
    split = [__remove_wrapping_quotes(item) for item in split]
    return split


def __remove_wrapping_quotes(item):
    match = re.match(r'(-+\w+=)*(?:(")[^"]*("))+', item, re.IGNORECASE)
    if match:
        return re.sub(r'"', "", item)
    match = re.match(r"(-+\w+=)*(?:(')[^']*('))+", item, re.IGNORECASE)
    if match:
        return re.sub(r"'", "", item)
    return item



import os
from urllib.request import Request


def add_github_access_token(request: Request) -> Request:
    token = os.getenv("GITHUB_ACCESS_TOKEN")
    if token:
        request.add_header("Authorization", f"token {token}")

    return request



import argparse
import sys
from typing import List, get_origin

from pydantic import TypeAdapter
from pydantic_core import PydanticUndefined

from hexagon.support.input.args import (
    CliArgs,
    ARGUMENT_KEY_PREFIX,
    OptionalArg,
    PositionalArg,
)
from hexagon.support.input.args.field_reference import FieldReference
from hexagon.typing import should_support_multiple_args, field_type_information


# noinspection PyProtectedMember
class HexagonFormatter(argparse.RawDescriptionHelpFormatter):
    pass


def cli_arg(cli_arguments, index):
    return (
        cli_arguments[index] if cli_arguments and len(cli_arguments) > index else None
    )


def parse_cli_args(args=None, model=CliArgs, **kwargs):
    if args is None:
        args = sys.argv[1:]

    _parser = init_arg_parser(model, **kwargs)

    known_args, extra = _parser.parse_known_args(args)

    if __no_known_args(known_args) and __arg_any_of(args, ["-h", "--help"]):
        return model(show_help=True, total_args=0)

    if __no_known_args(known_args) and __arg_any_of(extra, ["-v", "--version"]):
        return model(show_version=True, total_args=0)

    count, extra_args = __guess_optional_keys(extra)
    data = vars(known_args)
    data.update(
        {
            "raw_extra_args": extra,
            "extra_args": extra_args if extra_args else None,
            "total_args": len(list(x for x in known_args.__dict__.values() if x))
            + count,
        }
    )
    return model(**{k: v for k, v in data.items() if v is not None})


def __arg_any_of(args, values):
    return any([a in values for a in args])


def __no_known_args(known_args):
    return all(not x for x in known_args.__dict__.values())


def init_arg_parser(
    model, fields=None, prog=None, description=None, add_help=False, epilog=None
):
    __p = argparse.ArgumentParser(
        prog=prog or "hexagon",
        description=description or "Hexagon CLI",
        add_help=add_help,
        epilog=epilog,
        formatter_class=HexagonFormatter,
    )
    if sys.version_info < (3, 8):
        __polyfill_extend_action(__p)

    for name, field in model.model_fields.items():
        if fields and name not in fields:
            continue
        if get_origin(field.annotation) in [PositionalArg, OptionalArg]:
            __add_parser_argument(__p, FieldReference(name, field))
    return __p


def __polyfill_extend_action(__p):
    class ExtendAction(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            items = getattr(namespace, self.dest) or []
            items.extend(values)
            setattr(namespace, self.dest, items)

    __p.register("action", "extend", ExtendAction)


def __add_parser_argument(parser, field: FieldReference):
    reprs = field.info.annotation.cli_repr(field)
    nargs, action, constant_default, is_bool = __config_base_on_type(field.info)

    default = None if field.info.default is PydanticUndefined else field.info.default
    parser.add_argument(
        *[r for r in reprs if r],
        nargs=nargs,
        action=action,
        const=constant_default,
        help=f"{field.info.description or field.name} (default: {default})",
    )
    if is_bool:
        parser.add_argument(
            *[__bool_negated_key(r.replace("-", "")) for r in reprs if r],
            action="store_const",
            dest=field.name,
            const="false",
            help=f"Disable {field.name}",
        )


def __config_base_on_type(field):
    """
    Return a tuple with the number of arguments and the action to be taken
    if the field is any kind of collection, nargs should be * and the action is "extend"
    :param field:
    :return:
    """
    field_type, _, __ = field_type_information(field)
    constant_default = "true" if field_type.is_bool else None
    return (
        (
            "*",
            "extend",
            constant_default,
            field_type.is_bool,
        )
        if get_origin(field.annotation) == OptionalArg
        and should_support_multiple_args(field)
        else (
            "?",
            "store",
            constant_default,
            field_type.is_bool,
        )
    )


def __guess_optional_keys(extra: List[str]):
    """
    Return a dict by guessing optional keys from a list of arguments, e.g.:
    ['--number', '123', '--name', 'John', '--name', 'Doe'] -> {'number': 123,'name': ['John', 'Doe']}
    ['123'] -> {'0': 123}
    ['123', '--letter', 'A'] -> {'0': 123, 'letter': 'A'}
    :param extra:
    :return:
    """
    if not extra:
        return 0, None

    result = __args_to_key_vals(extra)
    return len(result), __group_by_key_appending(result)


def __args_to_key_vals(extra):
    result = []
    i = 0
    arg_index = 0
    while i < len(extra):
        current, next_ = extra[i], extra[i + 1] if i + 1 < len(extra) else None
        if current.startswith(ARGUMENT_KEY_PREFIX):
            if "=" in current:
                key, value = current[2:].split("=")
                result.append({key: __adapt_value_type(value)})
                i += 1
            elif not next_ or next_.startswith(ARGUMENT_KEY_PREFIX):
                result.append({current[2:]: True})
                i += 1
            else:
                result.append({current[2:]: __adapt_value_type(next_)})
                i += 2
        else:
            result.append({str(arg_index): __adapt_value_type(current)})
            i += 1
            arg_index += 1
    return result


def __group_by_key_appending(result):
    d = {}
    for r in result:
        for k, v in r.items():
            if k in d:
                if isinstance(d[k], list):
                    d[k].append(v)
                else:
                    d[k] = [d[k], v]
            else:
                d[k] = v
    return d


def __bool_negated_key(name: str):
    return f"{ARGUMENT_KEY_PREFIX * 2}no-{name}"


def __adapt_value_type(value: str):
    result = __try_to_parse_type(value, bool)
    if result is not None:
        return result
    return __try_to_parse_type(value, int) or __try_to_parse_type(value, float) or value


def __try_to_parse_type(value: str, target: type):
    adapter = TypeAdapter(target)
    try:
        return adapter.validate_strings(value)
    except ValueError:
        return None



import builtins
import gettext
import os
import site
import sys
from typing import List

LOCALEDIR = os.environ.get("HEXAGON_LOCALES_DIR", None)

# site.USER_BASE is usually ~/.local (this is used when pip WAS NOT installed globally)
LOCAL_LOCALEDIR = os.path.join(site.USER_BASE, "locales")

# sys.prefix is usually /usr/local (this is used when pip WAS installed globally)
SYSTEM_LOCALEDIR = os.path.join(sys.prefix, "locales")

DEFAULT_LANGUAGE = "en"


def install():
    """
    gettext translations are a PAIN, seriously considering of dropping it.

    Any-who, this function will install the translations for the current process.
    It will try to load the translations from the known paths, and fallback to the default ones.
    If nothing works we'll create a GNUTranslations object ourselves.
    """
    if getattr(builtins, "_", None):
        return
    el = (
        _load_first_match_from_dirs(
            [LOCALEDIR] if LOCALEDIR else [LOCAL_LOCALEDIR, SYSTEM_LOCALEDIR]
        )
        or _load_gettext_default()
        or _load_from_local()
    )

    el.install()


def _load_first_match_from_dirs(paths: List[str]):
    for path in paths:
        el = __try_translation(localedir=path)
        if el:
            return el
    for path in paths:
        el = __try_translation(localedir=path, languages=[DEFAULT_LANGUAGE])
        if el:
            return el
    return None


def _load_gettext_default():
    """
    gettext default path is os.path.join(sys.base_prefix, 'share', 'locale')
    for some reason when installing with pip locales are not copied there.
    """
    return __try_translation() or __try_translation(languages=[DEFAULT_LANGUAGE])


def _load_from_local():
    """
    If everything fails, we'll create a GNUTranslations object ourselves.
    File en.py is actually an .mo and is copied by .github/scripts/i18n/build.sh .
    We use .py, so it's packaged with everything else without any modifications to setup.

    Cheap tricks, I know.
    """
    return gettext.GNUTranslations(
        open(os.path.join(os.path.dirname(__file__), "en.py"), mode="rb")
    )


def __try_translation(**kwargs):
    try:
        return gettext.translation("hexagon", **kwargs)
    except FileNotFoundError:
        return None



from functools import reduce
from typing import List

from hexagon.runtime.update.changelog import ChangelogVersionEntry, ChangelogEntry

CHANGELOG_TYPE_FEATURE = "Feature"
CHANGELOG_TYPE_FIX = "Fix"
CHANGELOG_TYPE_DOCUMENTATION = "Documentation"

CHANGELOG_ENTRY_TYPE_ORDER_MAP = {
    CHANGELOG_TYPE_FEATURE: 2,
    CHANGELOG_TYPE_FIX: 1,
    CHANGELOG_TYPE_DOCUMENTATION: 0,
}


def __reducer(acc: List[ChangelogEntry], version: ChangelogVersionEntry):
    acc.extend(version.entries)
    return acc


def format_entries(entries: List[ChangelogVersionEntry]):
    entries = [
        entry
        for entry in reduce(
            __reducer,
            entries,
            [],
        )
        if entry.type in CHANGELOG_ENTRY_TYPE_ORDER_MAP.keys()
    ]
    entries.sort(key=lambda e: CHANGELOG_ENTRY_TYPE_ORDER_MAP[e.type], reverse=True)
    return entries



import time

from hexagon.domain.env import Env
from hexagon.domain.hooks.execution import ToolExecutionData
from hexagon.domain.tool import ActionTool
from hexagon.support.hooks import HexagonHooks


def execution_hook(func):
    """
    Decorator to add a hook to a tool execution, it will be called before and after the tool execution.

    :return: a wrapper for the decorated function
    """

    def wrapper(tool: ActionTool, env_args, env: Env, args):
        start = time.time()
        result = func(tool, env_args, env, args)
        HexagonHooks.tool_executed.run(
            ToolExecutionData(
                tool=tool, tool_env_args=env_args, duration=(time.time() - start)
            )
        )
        return result

    return wrapper



import os
import sys
from typing import List, Optional, Tuple, Union

from pydantic import BaseModel

from hexagon.domain.cli import Cli
from hexagon.domain.env import Env
from hexagon.domain.hexagon_error import ListHexagonError
from hexagon.domain.tool import (
    ActionTool,
    GroupTool,
    Tool,
    ToolType,
    ToolGroupConfigFile,
)
from hexagon.runtime.yaml import write_file, read_file, load_model


class ConfigFile(BaseModel):
    cli: Cli
    envs: List[Env]
    tools: List[Union[ActionTool, GroupTool]]


class GroupFileNotFoundError(ListHexagonError):
    def __init__(self, group_yaml_path: str):
        super().__init__(
            [
                _("error.domain.configuration.group_tool_file_not_found").format(
                    config_file_path=group_yaml_path
                )
            ]
        )


class Configuration:
    def __init__(self, path: str = None):
        self.project_path = path
        self.project_yaml = None
        self.custom_tools_path = None
        self.__yaml = None
        self.__config: Optional[ConfigFile] = None
        self.cwd_tools: List[Tool] = []

    def init_config(self, path: str) -> Tuple[Cli, List[Tool], List[Env]]:
        self.project_yaml = path
        self.project_path = os.path.dirname(self.project_yaml)

        self.__yaml = read_file(path)
        if not self.__yaml:
            return self.__initial_setup_config()

        self.__config = load_model(ConfigFile, self.__yaml, self.project_yaml)

        if self.__config.cli.custom_tools_dir:
            self.__register_custom_tools_path()

        self.__recursive_group_load(self.__config.tools)

        return self._configuration_tuple()

    def add_tools(self, tools: List[Tool]) -> Tuple[Cli, List[Tool], List[Env]]:
        self.cwd_tools += tools

        return self._configuration_tuple()

    def refresh(self) -> Tuple[Cli, List[Tool], List[Env]]:
        return self.init_config(self.project_yaml)

    def save(self) -> Tuple[Cli, List[Tool], List[Env]]:
        write_file(self.project_yaml, self.__yaml)
        return self._configuration_tuple()

    def add_tool(self, tool: Union[ActionTool, GroupTool]):
        self.__config.tools.append(tool)
        self.__yaml["tools"].append(
            tool.model_dump(mode="json", exclude_none=True, exclude_unset=True)
        )

    def update_custom_tools_path(self, value, comment=None, position=0):
        self.__config.cli.custom_tools_dir = value
        self.__yaml["cli"].insert(position, "custom_tools_dir", value, comment)
        self.__register_custom_tools_path()

    def _configuration_tuple(self):
        return (
            self.__config.cli,
            self.cwd_tools + self.__config.tools + self.__defaults,
            self.__config.envs,
        )

    @property
    def has_config(self):
        return self.__config is not None

    def __recursive_group_load(self, tools):
        for tool in tools:
            if tool.type == ToolType.group:
                if isinstance(tool.tools, str):
                    group_yaml_path = os.path.join(self.project_path, tool.tools)
                    group_config_yaml = read_file(group_yaml_path)

                    if not group_config_yaml:
                        raise GroupFileNotFoundError(group_yaml_path)

                    group_config = load_model(
                        ToolGroupConfigFile, group_config_yaml, group_yaml_path
                    )

                    tool.tools = group_config.tools
                self.__recursive_group_load(tool.tools)

    def __register_custom_tools_path(self):
        self.custom_tools_path = register_custom_tools_path(
            self.__config.cli.custom_tools_dir, self.project_path
        )

    @staticmethod
    def __initial_setup_config() -> Tuple[Cli, List[Tool], List[Env]]:
        def _tools():
            return [
                ActionTool(
                    name="install",
                    long_name=_("msg.domain.configuration.install_cli_long_name"),
                    description=_("msg.domain.configuration.install_cli_description"),
                    type=ToolType.hexagon,
                    action="hexagon.actions.internal.install_cli",
                ),
                ActionTool(
                    name="get-json-schema",
                    long_name=_("msg.domain.configuration.get_json_schema_long_name"),
                    description=_(
                        "msg.domain.configuration.get_json_schema_description"
                    ),
                    type=ToolType.hexagon,
                    action="hexagon.actions.internal.schema",
                ),
                ActionTool(
                    name="update-hexagon",
                    long_name=_("msg.domain.configuration.update_hexagon_long_name"),
                    type=ToolType.hexagon,
                    action="hexagon.actions.internal.update_hexagon",
                ),
            ]

        return (
            Cli(name="Hexagon", command="hexagon"),
            _tools(),
            [],
        )

    @property
    def __defaults(self):
        return [
            ActionTool(
                name="save-alias",
                long_name=_("msg.domain.configuration.save_alias_long_name"),
                type=ToolType.hexagon,
                action="hexagon.actions.internal.save_new_alias",
            ),
            ActionTool(
                name="replay",
                alias="r",
                long_name=_("msg.domain.configuration.replay_long_name"),
                type=ToolType.hexagon,
                action="hexagon.actions.internal.replay",
            ),
            ActionTool(
                name="create-tool",
                long_name=_("msg.domain.configuration.create_tool_long_name"),
                type=ToolType.hexagon,
                action="hexagon.actions.internal.create_new_tool",
            ),
            ActionTool(
                name="update-cli",
                long_name=_("msg.domain.configuration.update_cli_long_name"),
                type=ToolType.hexagon,
                action="hexagon.actions.internal.update_cli",
            ),
        ]


def register_custom_tools_path(path: str, realtive_to: str) -> str:
    absolute_path = os.path.abspath(os.path.join(realtive_to, path))
    sys.path.append(absolute_path)
    return absolute_path



import webbrowser

from hexagon.support.output.printer import log


def main(tool, env, env_args, cli_args):
    url = env_args
    log.result(_("msg.actions.external.open_link.result").format(url=url))
    webbrowser.open(url)



from hexagon.domain.env import Env
from hexagon.domain.hooks.execution import ToolExecutionParameters, ToolExecutionData
from hexagon.domain.hooks.wax import Selection
from hexagon.domain.tool import Tool
from hexagon.support.hooks.hook import Hook


class HexagonHooks:
    start = Hook[None]("start")
    tool_selected = Hook[Selection[Tool]]("tool_selected")
    env_selected = Hook[Selection[Env]]("env_selected")
    before_tool_executed = Hook[ToolExecutionParameters]("before_tool_executed")
    tool_executed = Hook[ToolExecutionData]("tool_executed")
    end = Hook[None]("end")



import os

from hexagon.domain import HEXAGON_CWD_TOOLS_FILE_NAME
from hexagon.domain.tool import OnlyToolsFile
from hexagon.runtime.options import Options
from hexagon.runtime.yaml import read_file, load_model


def collect_cwd_tools(options: Options):
    """
    Collects all tools from the current working directory and registers them in the configuration.
    :param options:
    """
    if options.cwd_tools_disabled:
        return None

    extra_tools_file = os.path.join(os.getcwd(), HEXAGON_CWD_TOOLS_FILE_NAME + ".yaml")
    if not os.path.exists(extra_tools_file):
        extra_tools_file = os.path.join(
            os.getcwd(), HEXAGON_CWD_TOOLS_FILE_NAME + ".yml"
        )

    if not os.path.exists(extra_tools_file):
        return None

    extra_tools_yaml = read_file(extra_tools_file)

    if not extra_tools_yaml:
        return None

    return load_model(OnlyToolsFile, extra_tools_yaml, extra_tools_file)
